import unittest

import gspread
import numpy as np

from neuralib.tools.gspread import *

# replace the following for testing
CFG_FILE = ...  # config filepath for gspread token (i.e., service_account.json)
EMAIL_ADDRESS = ...  # your email address


class TestSpreadSheet(unittest.TestCase):
    sh: GoogleSpreadSheet

    @classmethod
    def setUpClass(cls):
        """create test spreadsheet"""
        gc = gspread.service_account(filename=CFG_FILE)
        sh = gc.create('test_gspread')
        sh.share(EMAIL_ADDRESS, perm_type='user', role='writer')
        sh.add_worksheet('A', rows=10, cols=20)
        sh.add_worksheet('B', rows=20, cols=30)
        sh.del_worksheet(sh.worksheet('Sheet1'))

        #
        cls.sh = GoogleSpreadSheet('test_gspread', service_account_path=CFG_FILE)

    def test_len(self):
        self.assertEqual(len(self.sh), 2)

    # noinspection PyTypeChecker
    def test_contain(self):
        self.assertIn('A', self.sh)
        self.assertNotIn('C', self.sh)

    def test_title(self):
        self.assertEqual('test_gspread', self.sh.title)

    def test_worksheet_list(self):
        self.assertListEqual(['A', 'B'], self.sh.worksheet_list)

    @classmethod
    def tearDownClass(cls):
        cls.sh._client.del_spreadsheet(cls.sh._sheet.id)


class TestWorkSheetCust(unittest.TestCase):
    ws: GoogleWorkSheet

    @classmethod
    def setUpClass(cls):
        cls.ws = GoogleWorkSheet.of('Test_YWAnalysis', 'apcls_tac', service_account_path=CFG_FILE)

    def test_title(self):
        self.assertEqual('apcls_tac', self.ws.title)

    def test_headers(self):
        expect = [
            'Data', 'region', 'pair_wise_group', 'dark end', 'ch2_num', 'depth', 'n_planes',
            'loc_MA', 'loc_MP', 'loc_LP', 'loc_LA', 'rotation', 'Notes', 'TODO', 'Gernal Notes'
        ]
        self.assertListEqual(self.ws.headers, expect)

    def test_index_list(self):
        expect = ['210315_YW006', '210401_YW006', '210402_YW006', '210409_YW006', '210402_YW008', '210407_YW008',
                  '210409_YW008', '210416_YW008', '210604_YW010', '210610_YW010', '#210513_YW017', '210514_YW017',
                  '210519_YW017', '211202_YW022', '211209_YW022', '211203_YW032', '211208_YW032', '211202_YW033',
                  '211208_YW033', '221018_YW048', '221019_YW048', '#TODO', '# BACKUP', '211029_YW022', '## TBA',
                  '210514_YW018', '210518_YW018', '210526_YW018', '####', '220518_YW040', '220520_YW040',
                  '220526_YW040']
        self.assertListEqual(self.ws.index_value, expect)

    def test_get_range_value(self):
        region_a1 = 'B3:B5'
        region_exp = ['aRSC', 'pRSC', 'pRSC']
        self.assertListEqual(self.ws.get_range_value(region_a1), region_exp)

    def test_get_range_value_2d(self):
        loc_a1 = 'H2:K3'
        loc_exp = ['0;-1.17', '0;-2.17', '-0.67;-2.17', '-0.67;-1.17', '0;-1.17', '0;-2.17', '-0.67;-2.17',
                   '-0.67;-1.17']
        self.assertListEqual(self.ws.get_range_value(loc_a1), loc_exp)

    def test_get_value(self):
        exp = ['0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '0', '-5.1',
               '-6.9']
        self.assertListEqual(self.ws.values('rotation'), exp)

    def test_row(self):
        f = self.ws._row
        self.assertEqual(f(0), 2)
        self.assertEqual(f('210315_YW006'), 2)
        self.assertListEqual(f(['210604_YW010', '210610_YW010']), [10, 11])
        np.testing.assert_array_equal(f(np.array([1, 3])), np.array([3, 5]))

    def test_col(self):
        self.assertEqual(self.ws._col('pair_wise_group'), 3)

    def test_get_cells(self):
        v = self.ws.get_cell(data='# test', head='pair_wise_group')
        self.assertEqual(v, '0.5')

        f = self.ws.get_cell(data='# test', head='pair_wise_group', value_render_option='FORMULA')
        self.assertEqual(f, '=100/200')


if __name__ == '__main__':
    unittest.main()



"""
RasterMap Wrapper
=================

.. seealso:: `<https://github.com/MouseLand/rastermap>`_
.. seealso:: `<https://colab.research.google.com/github/MouseLand/rastermap/blob/main/notebooks/rastermap_largescale.ipynb>`_
.. seealso:: `<https://colab.research.google.com/github/MouseLand/rastermap/blob/main/notebooks/rastermap_widefield.ipynb>`_



Example of 2P dataset pipeline
-------------------------------


.. code-block:: python

    from __future__ import annotations

    import dataclasses

    import numpy as np
    import rastermap.utils
    from rastermap import Rastermap, rastermap

    from neuralib.plot import plot_figure, ax_merge
    from neuralib.util.util_type import PathLike
    from neuralib.wrapper.rastermap import RasterOptions, RasterMapResult, UserCluster


**1. Prepare container data input for rastermap**

    `Dimension parameters`:

        N = number of neurons

        T = number of image pulse

        S = number of stimulation (optional)


.. code-block:: python

    @dataclasses.dataclass
    class BaseRasterMapInput2P:
        xy_pos: np.ndarray  # soma central position (2, N)
        neural_activity: np.ndarray  # neural activity (N, T)
        image_time: np.ndarray  # 2p imaging time (T,)

        # optional user-specific behavioral variables
        position: np.ndarray | None
        velocity: np.ndarray | None
        pupil_area: np.ndarray | None
        stimulation_epoch: np.ndarray  # (S, 2) with on-off stim time

**2. Prepare config dict**

.. code-block:: python

    DEFAULT_2P_RASTERMAP_OPT: RasterOptions = {
        'n_clusters': 20,
        'n_PCs': 128,
        'locality': 0.75,
        'time_lag_window': 5,
        'grid_upsample': 10,
    }


**3. Run Rastermap and get results**

.. code-block:: python

    def run_rastermap_2p(dat: BaseRasterMapInput2P,
                         suite2p_directory: PathLike,
                         ops: RasterOptions | None = None,
                         neuron_bins: int = 30,
                         **kwargs) -> RasterMapResult:
        if ops is None:
            ops = DEFAULT_2P_RASTERMAP_OPT

        model = Rastermap(
            n_clusters=ops['n_clusters'],
            n_PCs=ops['n_PCs'],
            locality=ops['locality'],
            time_lag_window=ops['time_lag_window'],
            grid_upsample=ops['grid_upsample'],
            **kwargs
        ).fit(dat.neural_activity)

        embedding = model.embedding
        isort = model.isort

        sn = rastermap.utils.bin1d(dat.neural_activity[isort], bin_size=neuron_bins, axis=0)

        # For fit gui launch behavior
        _pseudo_cluster: UserCluster = {
            'ids': np.arange(10),
            'slice': slice(0, 10),
            'binsize': neuron_bins,
            'color': np.array([194.59459459, 255., 0., 50.])
        }

        ret = RasterMapResult(
            filename=str(suite2p_directory / 'F.npy'),
            save_path=str(suite2p_directory),
            isort=isort,
            embedding=embedding,
            ops=ops,
            user_clusters=[_pseudo_cluster],
            super_neurons=sn
        )

        return ret

**4. Plotting cluster and soma location color map**

.. code-block:: python

    def plot_rastermap_2p(raster: RasterMapResult, output: PathLike | None = None):
        with plot_figure(output, 10, 20, gridspec_kw={'wspace': 1, 'hspace': 0.3}) as _ax:
            # heatmap
            ax = ax_merge(_ax)[:, :-1]
            ax.imshow(raster.super_neurons,
                      cmap="gray_r",
                      vmin=0,
                      vmax=0.8,
                      aspect="auto",
                      interpolation='none')
            ax.set(xlabel="time(s)", ylabel='superneurons')

            # soma colorbar
            ax = ax_merge(_ax)[:, -1:]
            ax.imshow(np.arange(0, raster.n_super)[:, np.newaxis], cmap="gist_ncar", aspect="auto")
            ax.axis("off")


.. code-block:: python

    def plot_rastermap_2p_soma(dat: BaseRasterMapInput2P,
                               raster: RasterMapResult,
                               output: PathLike | None = None):
        with plot_figure(output) as ax:
            ax.scatter(dat.xy_pos[0],
                       dat.xy_pos[1],
                       s=8, c=raster.embedding, cmap="gist_ncar", alpha=0.25)
            ax.invert_yaxis()
            ax.set(xlabel='X position(mm)', ylabel='Y position(mm)')
            ax.set_aspect('equal')





Example of Wfield imaging dataset pipeline
-------------------------------------------


**1. Prepare container data input for rastermap**

    `Dimension parameters`:

        W = image width

        H = image height

        T = number of image pulse

        C = number of components after SVD reduction

        S = number of stimulation (optional)

.. code-block:: python


    from __future__ import annotations

    import dataclasses

    import numpy as np
    import rastermap.utils
    from rastermap import Rastermap
    from scipy.stats import zscore

    from neuralib.plot import plot_figure, ax_merge
    from neuralib.util.util_type import PathLike


    @dataclasses.dataclass
    class BaseRasterMapInputWField:
        sequences: np.ndarray  # image sequences (T, W, H)

        # singular vector
        n_components: int = dataclasses.field(init=False, default=128)
        sv: np.ndarray | None = dataclasses.field(init=False)  # (C, ) singular values
        Vsv: np.ndarray | None = dataclasses.field(init=False)  # (T, C) right singular vector
        U: np.ndarray | None = dataclasses.field(init=False)  # left singular vector (W * H, C)

        @property
        def n_frames(self) -> int:
            return self.sequences.shape[0]

        @property
        def width(self) -> int:
            return self.sequences.shape[1]

        @property
        def height(self) -> int:
            return self.sequences.shape[2]

        @property
        def xpos(self) -> np.ndarray:
            x = np.arange(self.width)
            y = np.arange(self.height)
            return np.meshgrid(x, y)[0]

        @property
        def ypos(self) -> np.ndarray:
            x = np.arange(self.width)
            y = np.arange(self.height)
            return np.meshgrid(x, y)[1]

        def compute_singular_vector(self) -> None:
            from sklearn.decomposition import TruncatedSVD

            # to 2D voxel
            seq = self.sequences.reshape(self.n_frames, self.height * self.width).T
            seq = seq - np.mean(seq, axis=1, keepdims=True)

            svd = TruncatedSVD(n_components=self.n_components)
            Vsv = svd.fit_transform(seq.T)
            U = seq @ (Vsv / (Vsv ** 2).sum(axis=0) ** 0.5)
            U /= (U ** 2).sum(axis=0) ** 0.5

            self.sv = svd.singular_values_
            self.Vsv = Vsv
            self.U = U


**2. Prepare config dict**

.. code-block:: python

    DEFAULT_WFIELD_RASTER_OPT: RasterOptions = {
        'n_clusters': 100,
        'locality': 0.5,
        'time_lag_window': 10,
        'grid_upsample': 10
    }


**3. Run Rastermap and get results**

.. code-block:: python


    def run_rastermap_wfield(dat: BaseRasterMapInputWField,
                             ops: RasterOptions | None = None,
                             neuron_bins: int = 500,
                             **kwargs) -> RasterMapResult:
        if ops is None:
            ops = DEFAULT_WFIELD_RASTER_OPT

        ops['n_PCs'] = dat.n_components

        model = Rastermap(
            n_clusters=ops['n_clusters'],
            n_PCs=ops['n_PCs'],
            locality=ops['locality'],
            time_lag_window=ops['time_lag_window'],
            grid_upsample=ops['grid_upsample'],
            **kwargs
        ).fit(
            Usv=dat.U * dat.sv,  # left singular vectors weighted by the singular values
            Vsv=dat.Vsv  # right singular vectors weighted by the singular values
        )

        embedding = model.embedding
        isort = model.isort
        Vsv_sub = model.Vsv  # these are the PCs across time with the mean across voxels subtracted

        U_sn = rastermap.utils.bin1d(dat.U[isort], bin_size=neuron_bins, axis=0)  # bin over voxel axis
        sn = U_sn @ Vsv_sub.T
        sn = zscore(sn, axis=1)

        ret = RasterMapResult(
            filename=...,  # replace to user specific
            save_path=...,  # replace to user specific
            isort=isort,
            embedding=embedding,
            ops=ops,
            user_clusters=[],
            super_neurons=sn
        )

        return ret


**4. Plotting cluster and FOV color map**


.. code-block:: python

    def plot_rastermap_sort(raster: RasterMapResult,
                            output: PathLike | None = None):
        with plot_figure(output,
                         9, 20,
                         dpi=200,
                         gridspec_kw={'wspace': 1, 'hspace': 0.3}) as _ax:
            ax = ax_merge(_ax)[:, :-1]
            ax.imshow(raster.super_neurons,
                      cmap="gray_r",
                      vmin=0,
                      vmax=0.8,
                      aspect="auto",
                      interpolation='none')
            ax.set(xlabel="time(s)", ylabel='superneurons')

            # soma
            ax = ax_merge(_ax)[:, -1:]
            ax.imshow(np.arange(0, raster.n_super)[:, np.newaxis], cmap="gist_ncar", aspect="auto")
            ax.axis("off")


    def plot_raster_voxel(dat: BaseRasterMapInputWField,
                          raster: RasterMapResult,
                          output: PathLike | None = None):
        with plot_figure(output) as ax:
            ax.scatter(dat.xpos,
                       dat.ypos,
                       s=1, c=raster.embedding, cmap="gist_ncar", alpha=0.25)
            ax.invert_yaxis()
            ax.set(xlabel='X position (um)', ylabel='Y position')
            ax.set_aspect('equal')


"""

from .core import *



from __future__ import annotations

import abc
import typing
from typing import NamedTuple, Any, Optional, Generic, TYPE_CHECKING, Callable

from .annotation import *
from .literal import FOREIGN_POLICY, CONFLICT_POLICY

if TYPE_CHECKING:
    from .expr import SqlExpr

__all__ = [
    'Field', 'table_name', 'table_field_names', 'table_fields', 'table_field',
    'table_primary_fields',
    'UniqueConstraint', 'table_unique_fields', 'make_unique_constraint',
    'ForeignConstraint', 'table_foreign_fields', 'table_foreign_field', 'make_foreign_constrain',
    'CheckConstraint', 'table_check_fields', 'table_check_field', 'make_check_constraint',

]

T = typing.TypeVar('T')
F = typing.TypeVar('F')
missing = object()


class Field(NamedTuple):
    """
    A SQL column, a field in a table.
    """
    table: type
    """associated table"""

    name: str
    """column/field name"""

    raw_type: type
    """origin field type"""

    sql_type: type
    """column/field type. Should be supported by SQL"""

    f_value: Any = missing
    """default value."""

    not_null: bool = True
    """Is it not NULL."""

    annotated: list[Any] = ()

    @property
    def table_name(self) -> str:
        return table_name(self.table)

    @property
    def has_default(self) -> bool:
        if self.f_value is not missing:
            return True

        if (primary := self.get_primary()) is not None:
            return primary.auto_increment

        return False

    @property
    def is_primary(self) -> bool:
        return self.get_primary() is not None

    def get_primary(self) -> PRIMARY | None:
        for a in self.annotated:
            if a == PRIMARY:
                return PRIMARY()
            elif isinstance(a, PRIMARY):
                return a
        return None

    @property
    def is_unique(self) -> bool:
        return self.get_unique() is not None

    def get_unique(self) -> UNIQUE | None:
        for a in self.annotated:
            if a == UNIQUE:
                return UNIQUE()
            elif isinstance(a, UNIQUE):
                return a
        return None

    def get_annotation(self, annotation_type: type[T]) -> T | None:
        for a in self.annotated:
            if isinstance(a, annotation_type):
                return a
        return None


class Table(Generic[T], metaclass=abc.ABCMeta):
    """
    SQL table information.
    """

    table_type: type[T]
    """associated table"""

    table_name: str
    """name of the table."""

    @abc.abstractmethod
    def table_seq(self, instance: T) -> tuple[Any, ...]:
        """
        cast an instance as a tuple as SQL parameters.

        :param instance:
        :return:
        """
        pass

    @abc.abstractmethod
    def table_dict(self, instance: T, *, sql_type: bool = True) -> dict[str, Any]:
        """
        cast an instance cast as dict as SQL parameters.

        :param instance:
        :param sql_type: cast value as SQL type
        """
        pass

    @abc.abstractmethod
    def table_new(self, *args) -> T:
        """create an instance."""
        pass

    @property
    def table_field_names(self) -> list[str]:
        """list of the name for each field in the table."""
        return [it.name for it in self.table_fields]

    @property
    @abc.abstractmethod
    def table_fields(self) -> list[Field]:
        """list fields in the table."""
        pass

    @property
    def table_primary_fields(self) -> list[Field]:
        """list of primary field in the table."""
        return [field for field in self.table_fields if field.is_primary]

    def table_field(self, name: str) -> Field:
        """
        Get field by the name in the table.

        :param name:
        :return:
        :raise RuntimeError: no such field.
        """
        for field in self.table_fields:
            if field.name == name:
                return field
        raise RuntimeError(f'{self.table_name} no such field {name}')

    @property
    @abc.abstractmethod
    def table_unique_fields(self) -> list[UniqueConstraint]:
        """get a list of the unique constraint in the table."""
        pass

    @property
    @abc.abstractmethod
    def table_foreign_fields(self) -> list[ForeignConstraint]:
        """get a list of the foreign constraint in the table."""
        pass

    @property
    @abc.abstractmethod
    def table_check_fields(self) -> dict[Optional[str], CheckConstraint]:
        """get a dict of the field constraint in the table."""
        pass


def table_class(cls: type[T]) -> Table[T]:
    return getattr(cls, '_sql_table')


def table_name(table: type[T]) -> str:
    """the name of the table."""
    return table_class(table).table_name


class UniqueConstraint(NamedTuple):
    name: str
    """constraint name"""

    table: type
    """associated table"""

    fields: list[str]
    """associated fields"""

    conflict: CONFLICT_POLICY | None


class ForeignConstraint(NamedTuple):
    """SQL foreign constraint."""

    name: str
    """constraint name"""

    table: type
    """associated table"""
    fields: list[str]
    """associated fields"""

    foreign_table: type
    """referred foreign table"""
    foreign_fields: list[str]
    """referred foreign fields"""

    on_update: FOREIGN_POLICY
    on_delete: FOREIGN_POLICY


class CheckConstraint(NamedTuple):
    """SQL check constraint"""
    name: str
    """constraint name"""

    table: type
    """associated table"""
    field: Optional[str]
    """associated field's name."""
    expression: SqlExpr
    """checking expression"""


def table_field_names(table: type[T]) -> list[str]:
    """list of the name for each field in the table."""
    return table_class(table).table_field_names


def table_fields(table: type[T]) -> list[Field]:
    """list fields in the table."""
    return table_class(table).table_fields


def table_field(table: type[T], field: str) -> Field:
    """
    Get field by the name in the table.

    :param table:
    :param field:
    :return:
    :raise RuntimeError: no such field.
    """
    return table_class(table).table_field(field)


def table_primary_fields(table: type[T]) -> list[Field]:
    """list of the name for each primary field in the table."""
    return table_class(table).table_primary_fields


def table_unique_fields(table: type[T]) -> list[UniqueConstraint]:
    """list of the name for each unique field in the table."""
    return table_class(table).table_unique_fields


def table_foreign_fields(table: type[T]) -> list[ForeignConstraint]:
    """get a list of the foreign constraint in the table."""
    return table_class(table).table_foreign_fields


@typing.overload
def table_foreign_field(table: Callable) -> ForeignConstraint | None:
    pass


@typing.overload
def table_foreign_field(table: type[T], target: type[F]) -> ForeignConstraint | None:
    pass


def table_foreign_field(table: type[T] | Callable, target: type[F] = None) -> ForeignConstraint | None:
    """
    get the foreign constraint in the table that refer to the target table.

    :param table: table or a foreign constraint function/property (the function decorated by @foreign)
    :param target: refer table
    :return: foreign constraint.
    """
    if isinstance(table, type) and isinstance(target, type):
        for constraint in table_foreign_fields(table):
            if constraint.foreign_table == target:
                return constraint
        return None
    elif target is None:
        constraint = getattr(table, '_sql_foreign', None)
        if isinstance(constraint, ForeignConstraint):
            return constraint
        return None
    else:
        raise TypeError(repr(table))


def table_check_fields(table: type[T]) -> dict[Optional[str], CheckConstraint]:
    """get a dict of the field constraint in the table."""
    return table_class(table).table_check_fields


def table_check_field(table: type[T], field: Optional[str]) -> Optional[CheckConstraint]:
    """get the check constrain of a field in the table."""
    return table_class(table).table_check_fields.get(field, None)


def make_foreign_constrain(table: Table,
                           prop: callable,
                           fields: list,
                           update: FOREIGN_POLICY,
                           delete: FOREIGN_POLICY) -> ForeignConstraint:
    from .expr import SqlField

    foreign_fields = []

    if len(fields) == 0:
        raise RuntimeError('empty fields')
    elif isinstance(fields[0], str):
        if not all([isinstance(it, str) for it in fields]):
            raise TypeError()

        foreign_table = table.table_type
        foreign_fields.extend(fields)
    elif len(fields) == 1 and isinstance(fields[0], type):
        foreign_table = fields[0]
        foreign_fields = [it.name for it in table_primary_fields(foreign_table)]
    else:
        foreign_table = None

        for field in fields:
            if isinstance(field, SqlField):
                if foreign_table is None:
                    foreign_table = field.table
                elif foreign_table != field.table:
                    raise RuntimeError()

                foreign_fields.append(field.name)
            else:
                raise TypeError()

    ret = prop(table.table_type)
    if not isinstance(ret, tuple):
        ret = [ret]

    self_fields = []
    for field in ret:
        if isinstance(field, SqlField):
            if table.table_type != field.table:
                raise RuntimeError()
            self_fields.append(field.name)
        else:
            raise TypeError()

    return ForeignConstraint(prop.__name__, table.table_type, self_fields, foreign_table, foreign_fields, update, delete)


def make_check_constraint(table: Table, prop: callable, field: str | None) -> CheckConstraint:
    from .expr import wrap
    ret = wrap(prop(table.table_type))
    return CheckConstraint(prop.__name__, table.table_type, field, ret)


def make_unique_constraint(table: Table, prop: callable, conflict: CONFLICT_POLICY = None) -> UniqueConstraint:
    from .expr import SqlField

    ret = prop(table.table_type)
    if not isinstance(ret, tuple):
        ret = [ret]

    fields = []
    for field in ret:
        if isinstance(field, SqlField):
            if table.table_type != field.table:
                raise RuntimeError()
            fields.append(field.name)
        else:
            raise TypeError()

    return UniqueConstraint(prop.__name__, table.table_type, fields, conflict)



"""
BrainRender Wrapper
===================

:author:
    Yu-Ting Wei

This module provide a CLI-based brainrender wrapper
See detail in the https://brainglobe.info/documentation/brainrender/index.html.
The wrapper provide three main usage cases, and can be run as command line once the package installed

Region reconstruction (area mode)
---------------------------------------

Plot brain regions

Example of reconstruct the Visual Cortex ::

    brender area -R VISal,VISam,VISl,VISli,VISp,VISpl,VISpm,VISpor --camera top


|brender area|


See the available options use `-h` option ::

    brender area -h



ROI reconstruction (roi mode)
---------------------------------------

Plot brain regions with ROIs label

Example of reconstruct ROIs in the Somatosensory Cortex for ipsilateral hemisphere(assume right hemisphere)::

    brender roi -F <CSV_FILE>


|brender roi|



CSV FILE example (auto transformed coordinates space from allen to brainrender)::

    ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┬─────────┬────────┬───────────────────────────┬──────────────┬────────┬────────────┬────────────┬────────────┬────────────┬────────────┬───────────┐
    │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex ┆ channel ┆ source ┆ abbr                      ┆ acronym_abbr ┆ hemi.  ┆ merge_ac_0 ┆ merge_ac_1 ┆ merge_ac_2 ┆ merge_ac_3 ┆ merge_ac_4 ┆ family    │
    │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     ┆ ---     ┆ ---    ┆ ---                       ┆ ---          ┆ ---    ┆ ---        ┆ ---        ┆ ---        ┆ ---        ┆ ---        ┆ ---       │
    │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     ┆ str     ┆ str    ┆ str                       ┆ str          ┆ str    ┆ str        ┆ str        ┆ str        ┆ str        ┆ str        ┆ str       │
    ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╪═════════╪════════╪═══════════════════════════╪══════════════╪════════╪════════════╪════════════╪════════════╪════════════╪════════════╪═══════════╡
    │ Ectorhinal area/Layer 5           ┆ ECT5    ┆ -3.03       ┆ 4.34        ┆ -4.5        ┆ 377     ┆ gfp     ┆ aRSC   ┆ Ectorhinal area           ┆ ECT          ┆ contra ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ISOCORTEX │
    │ Perirhinal area layer 6a          ┆ PERI6a  ┆ -3.03       ┆ 4.42        ┆ -4.37       ┆ 372     ┆ gfp     ┆ aRSC   ┆ Perirhinal area           ┆ PERI         ┆ contra ┆ PERI       ┆ PERI       ┆ PERI       ┆ PERI       ┆ PERI       ┆ ISOCORTEX │
    │ …                                 ┆ …       ┆ …           ┆ …           ┆ …           ┆ …       ┆ …       ┆ …      ┆ …                         ┆ …            ┆ …      ┆ …          ┆ …          ┆ …          ┆ …          ┆ …          ┆ …         │
    │ Ventral auditory area layer 6a    ┆ AUDv6a  ┆ -2.91       ┆ 3.52        ┆ 4.46        ┆ 156     ┆ rfp     ┆ pRSC   ┆ Ventral auditory area     ┆ AUDv         ┆ ipsi   ┆ AUD        ┆ AUD        ┆ AUD        ┆ AUD        ┆ AUDv       ┆ ISOCORTEX │
    │ Ectorhinal area/Layer 6a          ┆ ECT6a   ┆ -2.91       ┆ 4.14        ┆ 4.47        ┆ 378     ┆ rfp     ┆ pRSC   ┆ Ectorhinal area           ┆ ECT          ┆ ipsi   ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ECT        ┆ ISOCORTEX │
    │ Temporal association areas layer… ┆ TEa5    ┆ -2.91       ┆ 4.02        ┆ 4.55        ┆ 365     ┆ rfp     ┆ pRSC   ┆ Temporal association area ┆ TEa          ┆ ipsi   ┆ TEa        ┆ TEa        ┆ TEa        ┆ TEa        ┆ TEa        ┆ ISOCORTEX │
    └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┴─────────┴────────┴───────────────────────────┴──────────────┴────────┴────────────┴────────────┴────────────┴────────────┴────────────┴───────────┘

See how to create the csv after ccf pipeline

.. code-block:: python

    from neuralib.atlas.ccf.classifier import RoiClassifier
    from neuralib.atlas.ccf.core import AbstractCCFDir
    root = ...
    ccf_dir = AbstractCCFDir(root, with_overlap_sources=False)
    classifier = RoiClassifier(ccf_dir, plane='coronal')
    df = classifier.parsed_df

Example ccf data folder structure in (:class:`~neuralib.atlas.ccf.core.AbstractCCFDir()`)



See the available options use ``-h`` option ::

    brender roi -h



Probe reconstruction (probe mode)
---------------------------------------

Reconstruct the probe (or shanks) in accordance with trajectory labeling (e.g., DiI, DiO or lesion track...)

Prepare CSV file from ccf pipeline::

    ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┐
    │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex │
    │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     │
    │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     │
    ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╡
    │ Primary visual area layer 6a      ┆ VISp6a  ┆ -3.81       ┆ 1.92        ┆ -3.12       ┆ 191     │
    │ optic radiation                   ┆ or      ┆ -4.08       ┆ 2.33        ┆ -3.12       ┆ 1217    │
    │ Posterolateral visual area layer… ┆ VISpl6a ┆ -4.28       ┆ 2.29        ┆ -3.12       ┆ 198     │
    │ Posterolateral visual area layer… ┆ VISpl5  ┆ -4.52       ┆ 2.17        ┆ -3.12       ┆ 197     │
    │ Subiculum                         ┆ SUB     ┆ -3.93       ┆ 4.36        ┆ -3.3        ┆ 536     │
    │ Entorhinal area medial part dors… ┆ ENTm5   ┆ -4.19       ┆ 4.39        ┆ -3.3        ┆ 515     │
    │ Entorhinal area medial part dors… ┆ ENTm2   ┆ -4.44       ┆ 4.39        ┆ -3.3        ┆ 510     │
    │ Entorhinal area medial part dors… ┆ ENTm1   ┆ -4.66       ┆ 4.29        ┆ -3.3        ┆ 509     │
    └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┘

- Row number equal to shank numbers * 2 (in the example, use the 4 shanks NeuroPixel probe), 2 points indicate the most dorsal & ventral detected signals on the serial brain slices

- Use ``-P`` to specify the slice cutting orientation {coronal,sagittal,transverse}. If multiple shanks were inserted along the AP axis, assume do the sagittal plane, if inserted along the ML axis, then assume the coronal plane

- Assume shanks are not bended to perform interpolation based on the insert depth ``-D``



Example of Above csv file for targeting the left Entorhinal cortex (ENT) using 4 shanks NeuroPixel probe::

    brender probe -F <CSV_FILE> -D 3000 -P sagittal -R ENT -H left




|brender probe|




See the available options use ``-h`` option ::

    brender probe -h




.. |brender area| image:: ../_static/brender_area.png
.. |brender roi| image:: ../_static/brender_roi.png
.. |brender probe| image:: ../_static/brender_probe.png


"""

from .core import *
from .probe_rst import *
from .roi_rst import *



import unittest


class TestUtilFunc(unittest.TestCase):

    def test_check_attrs_in_clazz(self):
        from neuralib.util.util_clazz import check_attrs_in_clazz

        class Parent:
            a: int
            b: bool
            c: str

        class Child(Parent):
            d: float
            e: dict

        self.assertTrue(check_attrs_in_clazz(Parent, 'a'))
        self.assertFalse(check_attrs_in_clazz(Parent, 'd'))
        self.assertTrue(check_attrs_in_clazz(Child, 'a'))
        self.assertTrue(check_attrs_in_clazz(Child, 'e'))

    def test_key_from_value(self):
        from neuralib.util.utils import key_from_value

        dy = dict(
            a=[1, 2, 3],
            b=5,
            c=(4, 5, 3),
            d=30.5
        )

        self.assertEqual(key_from_value(dy, 3), ['a', 'c'])
        self.assertEqual(key_from_value(dy, 5), ['b', 'c'])
        self.assertEqual(key_from_value(dy, 30.5), 'd')

        with self.assertRaises(KeyError):
            key_from_value(dy, 100)



from bokeh.model import Model
from bokeh.models import Div

from neuralib.bokeh_model import View, BokehServer


class Top(View):

    @property
    def title(self) -> str:
        return 'A'

    def setup(self) -> Model:
        from bokeh.layouts import column
        return column(Div(text='root'),
                      Div(text='<a href="pool" >pool</a>'))


class Pool(View):

    @property
    def title(self) -> str:
        return 'B'

    def setup(self) -> Model:
        from bokeh.layouts import column
        return column(Div(text='pool'))


if __name__ == '__main__':
    VIEW_TOP = Top()
    VIEW_POOL = Pool()

    BokehServer().start({'/': VIEW_TOP, '/pool': VIEW_POOL})



from html import escape
from typing import Optional

from bokeh.model import Model
from bokeh.models import Select, Div

from neuralib.bokeh_model import BokehServer, View
from .model import *

__all__ = ['AllView']


class AllView(View):
    goto_btn: Div
    content: Div
    select_animal: Select
    select_date: Select

    def __init__(self, single=False):
        """

        :param single: disable jump to pool page.
        """
        self.single = single

    @property
    def title(self) -> str:
        return 'ALL'

    def setup(self) -> Model:
        animals = list_animals()
        self.select_animal = Select(
            title='Animal',
            value='',
            options=['', *animals]
        )
        self.select_animal.on_change('value', self.on_select_animal)

        self.select_date = Select(
            title='Date',
            value='',
            options=[]
        )
        self.select_date.on_change('value', self.on_select_date)

        self.goto_btn = Div(text='')

        self.content = Div(text='')

        from bokeh.layouts import row, column
        return column(
            row(
                self.select_animal,
                self.select_date,
                self.goto_btn,
            ),
            self.content
        )

    def update(self):
        self.build_goto_link(None)

    def on_select_animal(self, attr: str, old: str, value: str):
        if len(value):
            self.select_date.options = ['', *list_date(value)]
            self.build_goto_link(value)
        else:
            self.build_goto_link(None)

    def on_select_date(self, attr: str, old: str, value: str):
        if len(value):
            animal = self.select_animal.value
            self.content.text = f'show animal {animal} on date {value}'

    def build_goto_link(self, animal: Optional[str]):
        if animal is None or self.single:
            self.goto_btn.text = f"""
<a type=button 
    class="bk bk-input"
>(Select one animal)</a>
"""
        else:
            self.goto_btn.text = f"""
<a type=button 
    class="bk bk-input"
    href="/pool?animal={animal}"
>{escape(animal)}</a>
"""


def main():
    BokehServer().start(AllView(single=True))


if __name__ == '__main__':
    main()



import os
import subprocess

__all__ = ['NERUOLIB_VERSION',
           'get_commit_hash']


def get_commit_hash() -> str:
    codedir = os.path.dirname(os.path.abspath(__file__))
    try:
        commit_ref = (
            subprocess
            .check_output(['git', 'rev-parse',
                           '--verify', 'HEAD',
                           '--short'], cwd=codedir)
            .decode()
            .strip('\n')
        )

    except subprocess.CalledProcessError:
        commit_ref = ''

    return commit_ref


NERUOLIB_VERSION = get_commit_hash()



from __future__ import annotations

import sys
from typing import TypedDict

__all__ = [
    'KeyMapping',
    #
    'get_keymapping',
    'find_key_from_value'
]


class KeyMapping(TypedDict, total=False):
    """For controlling the keyboard in different OS"""
    escape: int
    backspace: int
    # space: int  # labeller specific printable for notes
    enter: int

    left_square_bracket: int  # [
    right_square_bracket: int  # ]

    left: int
    right: int
    up: int
    down: int

    plus: int  # +
    minus: int  # -
    equal: int  # =


#
COMMON_KEYMAPPING: KeyMapping = {
    'escape': 27,
    # 'space': 32,
    'enter': 13,

    'left_square_bracket': 91,
    'right_square_bracket': 93,

    # ord
    'plus': ord('+'),
    'minus': ord('-'),
    'equal': ord('=')
}

WIN_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 8,
    'left': 2424832,
    'right': 2555904,
    'up': 2490368,
    'down': 2621440
}

MAC_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 127,
    'left': 2,
    'right': 3,
    'up': 0,
    'down': 1

}

LINUX_KEYMAPPING: KeyMapping = {
    **COMMON_KEYMAPPING,
    'backspace': 8,
    'left': 81,
    'right': 83,
    'up': 82,
    'down': 84
}


def get_keymapping() -> KeyMapping:
    p = sys.platform
    if p in ('linux', 'linux2'):
        return LINUX_KEYMAPPING
    elif p == 'darwin':
        return MAC_KEYMAPPING
    elif p == 'win32':
        return WIN_KEYMAPPING


def find_key_from_value(dy: KeyMapping, value: int) -> str | bool | None:
    ret = []
    for k, v in dy.items():
        if v == value:
            ret.append(k)

    if len(ret) == 0:
        return
    elif len(ret) == 1:
        return ret[0]
    else:
        raise False



from __future__ import annotations

import numpy as np

__all__ = ['grey2rgb']


def grey2rgb(im: np.ndarray,
             rgb: int | tuple[int, ...], *,
             out: np.ndarray | None = None):
    """
    Transfer grey scale image to rgb

    :param im: grey scale image (2d array)
    :param rgb: {0, 1, 2} indicates rgb, tuple represent merge color
    :param out: output array
    :return:
    """
    from neuralib.imglib.norm import get_percentile_value

    if im.ndim != 2:
        raise ValueError('input should be an grayscale 2d array')

    #
    if out is None:
        out = np.zeros((*im.shape, 3), float)
    else:
        if out.shape != (*im.shape, 3):
            raise ValueError

    #
    lb, ub = get_percentile_value(im, perc_interval=(10, 100))
    if ub == lb:
        raise ValueError("Cannot normalize with zero range in image.")

    #
    if isinstance(rgb, int):
        out[:, :, rgb] = (im - lb) / ub
    else:
        for i in rgb:
            out[:, :, i] = (im - lb) / ub

    return np.clip(out, 0, 1)



from typing import ClassVar

from cellpose import denoise

from .core import AbstractCellPoseOption, CellPoseEvalResult

__all__ = ['CellPoseEvalResult']


class CellPoseAPIOption(AbstractCellPoseOption):
    DESCRIPTION = 'run cell pose for directly call the api'

    RESTORE_TYPE: ClassVar[str] = 'denoise_cyto3'
    RESTORE_RATIO: ClassVar[float] = 1.0

    def run(self):
        if self.seg_result.exists():
            if self.napari_view:
                self.launch_napari()
            elif self.cellpose_gui:
                self.launch_cellpose_gui()
        else:
            res = self.eval()
            res.save_seg_file()

    def get_model(self) -> denoise.CellposeDenoiseModel:
        return denoise.CellposeDenoiseModel(gpu=True,
                                            model_type=self.model,
                                            restore_type=self.RESTORE_TYPE,
                                            chan2_restore=True)

    def eval(self) -> CellPoseEvalResult:
        img = self.load_gray_scale()
        channel_choose = [int(self.chan_seg), int(self.chan_nuclear)]
        model = self.get_model()
        masks, flows, styles, imgs_dn = model.eval(img,
                                                   channels=channel_choose,
                                                   diameter=self.diameter)
        return CellPoseEvalResult(
            str(self.file.with_suffix('')),
            img,
            self.diameter,
            channel_choose,
            masks,
            flows,
            styles,
            img_restore=imgs_dn,
            restore_type=self.RESTORE_TYPE,
            ratio=self.RESTORE_RATIO
        )


if __name__ == '__main__':
    CellPoseAPIOption().main()



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Literal

import nrrd
import numpy as np
import pandas as pd
import polars as pl

from neuralib.util.io import CCF_CACHE_DIRECTORY, ALLEN_SDK_DIRECTORY
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint

__all__ = [
    'DATA_SOURCE_TYPE',
    #
    'load_ccf_annotation',
    'load_ccf_template',
    'load_structure_tree',
    #
    'load_allensdk_annotation',

]

DATA_SOURCE_TYPE = Literal['ccf_annotation', 'ccf_template', 'allensdk_annotation']


# ===================== #
# AllenCCF Data Source #
# ===================== #

def _cache_nparray(url: str, file: PathLike) -> None:
    fprint(f'DOWNLOADING... {file.name} from {url}', vtype='io')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        content = download_with_tqdm(url)
        data = np.load(content, allow_pickle=True)
        np.save(file, data)


def load_ccf_annotation(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Annotation volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return:
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'annotation_volume_10um_by_index.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925493'
        _cache_nparray(url, file)

    return np.load(file, allow_pickle=True)


def load_ccf_template(output_dir: PathLike | None = None) -> np.ndarray:
    """
    Template volume file from AllenCCF pipeline

    .. seealso::

        https://github.com/cortex-lab/allenCCF

    :param output_dir: output directory for caching
    :return:
    """
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / 'template_volume_10um.npy'

    if not file.exists():
        url = 'https://figshare.com/ndownloader/files/44925496'
        _cache_nparray(url, file)

    return np.load(file, allow_pickle=True)


def load_structure_tree(version: Literal['2017', 'old'] = '2017',
                        output_dir: PathLike | None = None) -> pl.DataFrame:
    """
    Load structure tree dataframe

    :param version: {'2017', 'old'}
    :param output_dir: output directory for caching
    :return: structure tree dataframe

    ::

        ┌───────────┬──────────┬───────────────────────────────┬─────────┬──────────┬─────────────┬───────────────┬────────┬─────────────────────┬───────┬──────────┬─────────────┬───────────────────┬───────────────────┬─────────────────────────┬──────────────────────────────┬────────┬───────────┬──────────────────────┬──────────────┬───────────────────────────────┐
        │ id        ┆ atlas_id ┆ name                          ┆ acronym ┆ st_level ┆ ontology_id ┆ hemisphere_id ┆ weight ┆ parent_structure_id ┆ depth ┆ graph_id ┆ graph_order ┆ structure_id_path ┆ color_hex_triplet ┆ neuro_name_structure_id ┆ neuro_name_structure_id_path ┆ failed ┆ sphinx_id ┆ structure_name_facet ┆ failed_facet ┆ safe_name                     │
        │ ---       ┆ ---      ┆ ---                           ┆ ---     ┆ ---      ┆ ---         ┆ ---           ┆ ---    ┆ ---                 ┆ ---   ┆ ---      ┆ ---         ┆ ---               ┆ ---               ┆ ---                     ┆ ---                          ┆ ---    ┆ ---       ┆ ---                  ┆ ---          ┆ ---                           │
        │ i64       ┆ f64      ┆ str                           ┆ str     ┆ str      ┆ i64         ┆ i64           ┆ i64    ┆ f64                 ┆ i64   ┆ i64      ┆ i64         ┆ list[i64]         ┆ str               ┆ str                     ┆ str                          ┆ str    ┆ i64       ┆ i64                  ┆ i64          ┆ str                           │
        ╞═══════════╪══════════╪═══════════════════════════════╪═════════╪══════════╪═════════════╪═══════════════╪════════╪═════════════════════╪═══════╪══════════╪═════════════╪═══════════════════╪═══════════════════╪═════════════════════════╪══════════════════════════════╪════════╪═══════════╪══════════════════════╪══════════════╪═══════════════════════════════╡
        │ 997       ┆ -1.0     ┆ root                          ┆ root    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ -1.0                ┆ 0     ┆ 1        ┆ 0           ┆ [997]             ┆ FFFFFF            ┆ null                    ┆ null                         ┆ f      ┆ 1         ┆ 385153371            ┆ 734881840    ┆ root                          │
        │ 8         ┆ 0.0      ┆ Basic cell groups and regions ┆ grey    ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1           ┆ [997, 8]          ┆ BFDAE3            ┆ null                    ┆ null                         ┆ f      ┆ 2         ┆ 2244697386           ┆ 734881840    ┆ Basic cell groups and regions │
        │ …         ┆ …        ┆ …                             ┆ …       ┆ …        ┆ …           ┆ …             ┆ …      ┆ …                   ┆ …     ┆ …        ┆ …           ┆ …                 ┆ …                 ┆ …                       ┆ …                            ┆ …      ┆ …         ┆ …                    ┆ …            ┆ …                             │
        │ 65        ┆ 715.0    ┆ parafloccular sulcus          ┆ pfs     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1040.0              ┆ 3     ┆ 1        ┆ 1324        ┆ [997, 1024, … 65] ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1325      ┆ 771629690            ┆ 734881840    ┆ parafloccular sulcus          │
        │ 624       ┆ 926.0    ┆ Interpeduncular fossa         ┆ IPF     ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 1024.0              ┆ 2     ┆ 1        ┆ 1325        ┆ [997, 1024, 624]  ┆ AAAAAA            ┆ null                    ┆ null                         ┆ f      ┆ 1326      ┆ 1476705011           ┆ 734881840    ┆ Interpeduncular fossa         │
        │ 304325711 ┆ -2.0     ┆ retina                        ┆ retina  ┆ null     ┆ 1           ┆ 3             ┆ 8690   ┆ 997.0               ┆ 1     ┆ 1        ┆ 1326        ┆ [997, 304325711]  ┆ 7F2E7E            ┆ null                    ┆ null                         ┆ f      ┆ 1327      ┆ 3295290839           ┆ 734881840    ┆ retina                        │
        └───────────┴──────────┴───────────────────────────────┴─────────┴──────────┴─────────────┴───────────────┴────────┴─────────────────────┴───────┴──────────┴─────────────┴───────────────────┴───────────────────┴─────────────────────────┴──────────────────────────────┴────────┴───────────┴──────────────────────┴──────────────┴───────────────────────────────┘

    """
    #
    if output_dir is None:
        output_dir = CCF_CACHE_DIRECTORY
        if not CCF_CACHE_DIRECTORY.exists():
            CCF_CACHE_DIRECTORY.mkdir(exist_ok=True, parents=True)
    #
    if version == '2017':
        filename = 'structure_tree_safe_2017.csv'
    elif version == 'old':
        filename = 'structure_tree_safe.csv'
    else:
        raise ValueError('')

    #
    file = output_dir / filename
    if not file.exists():
        url = f'https://raw.githubusercontent.com/cortex-lab/allenCCF/master/{filename}'
        content = download_with_tqdm(url)
        df = pd.read_csv(content)
        df.to_csv(file)
        fprint(f'DOWNLOAD! {filename} in {output_dir}', vtype='io')

    ret = (pl.read_csv(file)
           .with_columns(pl.col('parent_structure_id').fill_null(-1))
           .with_columns(pl.col('structure_id_path')
                         .map_elements(lambda it: tuple(map(int, it[1:-1].split('/'))),
                                       return_dtype=pl.List(pl.Int64))))
    #
    if '2017' in Path(file).name:
        ret = ret.with_columns(pl.col('atlas_id').fill_null(-2))
    else:
        ret = ret.with_columns(pl.col('name').alias('safe_name'))

    return ret


# ===================== #
# Allen SDK Data Source #
# ===================== #

def load_allensdk_annotation(resolution: int = 10,
                             output_dir: PathLike | None = None) -> np.ndarray:
    """
    Data Source directly from Allen Institute

    .. seealso::

        https://download.alleninstitute.org/informatics-archive/current-release/mouse_ccf/annotation/

    :param resolution: volume resolution in um. default is 10 um
    :param output_dir: output directory for caching
    :return:
    """
    if output_dir is None:
        output_dir = ALLEN_SDK_DIRECTORY
        if not ALLEN_SDK_DIRECTORY.exists():
            ALLEN_SDK_DIRECTORY.mkdir(exist_ok=True, parents=True)

    file = output_dir / f'annotation_{resolution}.nrrd'

    if not file.exists():
        from allensdk.api.queries.mouse_connectivity_api import MouseConnectivityApi
        mcapi = MouseConnectivityApi()
        version = MouseConnectivityApi.CCF_VERSION_DEFAULT

        mcapi.download_annotation_volume(version, resolution, file)

    return nrrd.read(file)[0]



from __future__ import annotations

from typing import Optional, Any, NamedTuple

from matplotlib.axes import Axes

__all__ = ['VennHandler',
           'plot_venn']


class VennHandler(NamedTuple):
    subset_a: int
    """whole number with condition a"""
    subset_b: int
    """whole number with condition b"""
    subset_overlap: int
    """whole number with condition a & b"""

    total_set: Optional[int] = None

    @property
    def chance_level(self) -> float:
        if self.total_set is None:
            raise ValueError('call with_total first')

        fa = self.subset_a / self.total_set
        fb = self.subset_b / self.total_set

        return fa * fb * 100

    def with_total(self, total: int) -> 'VennHandler':
        """total set number. should include the non-classified population"""
        return self._replace(total_set=total)

    def get_pure_number(self) -> tuple[int, ...]:
        a = self.subset_a - self.subset_overlap
        b = self.subset_b - self.subset_overlap
        return tuple([a, b, self.subset_overlap])

    def get_pure_fraction(self) -> tuple[float, ...]:
        a, b, o = self.get_pure_number()
        return tuple([a / self.total_set * 100,
                      b / self.total_set * 100,
                      o / self.total_set * 100])


def plot_venn(ax: Axes,
              vhandler: VennHandler,
              msg: Optional[list[Any]] = None,
              labels: tuple[str, str] = ('visual', 'place'),
              show_msg: bool = True):
    """
    three classified population venn diagram, and calculate the chance level

    :param ax: ``Axes``
    :param vhandler: VennHandler
    :param msg: msg show in title
    :param labels: label inside the venn
    :param show_msg: show fraction message
    :return:
    """
    from matplotlib_venn import venn2

    venn2(ax=ax, subsets=vhandler.get_pure_number(), set_labels=labels, set_colors=('pink', 'palegreen'))

    title = []
    if show_msg:
        vfrac, pfrac, ofrac = vhandler.get_pure_fraction()
        title.append(f'chance: {round(vhandler.chance_level, 2)}%')
        title.append(f'fraction: {round(vfrac, 2)}%, {round(ofrac, 2)}%, {round(pfrac, 2)}%')
        title.append(f'total: {vhandler.total_set}')

    if msg is not None:
        title.extend(map(str, msg))

    title = '\n'.join(title)

    ax.set_title(title)

    ax.set_axis_off()
    ax.set_xticks([])
    ax.set_yticks([])



from __future__ import annotations

import pickle
from typing import TypedDict, Generator, overload, final, Final, Literal

import attrs
import h5py
import numpy as np
from typing_extensions import TypeAlias, Self

from neuralib.util.cli_args import CliArgs
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import uglob

__all__ = [
    'TRACK_TYPE',
    'SVDVariables',
    #
    'KeyPoint',
    'FaceMapResult',
    'KeyPointTrack',
]

TRACK_TYPE = Literal['keypoints', 'pupil']


class PupilDict(TypedDict):
    """
    `Dimension parameters`:

        F: number pf frames
    """
    area: np.ndarray
    """(F,)"""
    com: np.ndarray
    """center of maze (F, 2)"""
    axdir: np.ndarray
    """(F, 2, 2)"""
    axlen: np.ndarray
    """(F, 2)"""
    area_smooth: np.ndarray
    """(F,)"""
    com_smooth: np.ndarray
    """(F, 2)"""


class RoiDict(TypedDict, total=False):
    rind: int
    rtype: str
    iROI: int
    ivid: int
    color: tuple[float, float, float]
    yrange: np.ndarray
    xrange: np.ndarray
    saturation: float
    pupil_sigma: float
    ellipse: np.ndarray
    yrange_bin: np.ndarray
    xrange_bin: np.ndarray


class SVDVariables(TypedDict, total=False):
    """SVD output from facemap

    .. seealso:: `<http://facemap.readthedocs.io/en/stable/outputs.html#roi-and-svd-processing>`_"""
    filenames: list[str]
    save_path: str
    Ly: list[int]
    Lx: list[int]
    sbin: int
    fullSVD: bool
    save_mat: bool
    Lybin: np.ndarray
    Lxbin: np.ndarray
    sybin: np.ndarray
    sxbin: np.ndarray
    LYbin: int
    LXbin: int
    avgframe: list[np.ndarray]
    avgmotion: list[np.ndarray]
    avgframe_reshape: np.ndarray
    avgmotion_reshape: np.ndarray
    motion: list[np.ndarray]
    motSv: list[np.ndarray]
    movSv: list[np.ndarray]
    motMask: list[int]
    movMask: list[int]
    motMask_reshape: list[int]
    movMask_reshape: list[int]
    motSVD: list[np.ndarray]
    movSVD: list[np.ndarray]
    pupil: list[PupilDict]
    running: list[np.ndarray]  # TODO check
    blink: list[np.ndarray]
    rois: list[RoiDict]  # TODO check
    sy: np.ndarray
    sx: np.ndarray


class KeyPointsMeta(TypedDict):
    """ Keypoint meta
    .. seealso:: `<https://facemap.readthedocs.io/en/stable/outputs.html#keypoints-processing>`_"""
    batch_size: int
    image_size: tuple[list[int], ...]
    bbox: tuple[int, ...]
    total_frames: int
    bodyparts: list[str]
    inference_speed: float


# ============== #
# FaceMap Result #
# ============== #

KeyPoint: TypeAlias = str


@final
class FaceMapResult:
    """facemap result container

    `Dimension parameters`:

        F = number of video frames

        K = number of keypoints

    """
    __slots__ = ('svd', 'meta', 'data', 'frame_time',
                 'track_type', 'with_keypoints')

    def __init__(
            self,
            svd: SVDVariables | None,
            meta: KeyPointsMeta | None,
            data: h5py.Group | None,
            track_type: TRACK_TYPE,
            with_keypoints: bool,
    ):
        """
        :param svd: SVD processing outputs
        :param meta: Optional for Keypoints processing (result)
        :param data: Optional for Keypoints processing (config)
        :param track_type: {'keypoints', 'pupil'}
        :param with_keypoints: if has keypoint tracking result
        """
        self.svd: Final[SVDVariables | None] = svd
        self.meta: Final[KeyPointsMeta | None] = meta
        self.data: Final[h5py.Group | None] = data

        self.track_type: Final[TRACK_TYPE] = track_type
        self.with_keypoints: Final[bool] = with_keypoints

    @classmethod
    def load(cls, directory: PathLike,
             track_type: TRACK_TYPE,
             *,
             file_pattern: str = '') -> Self:
        """
        Load the facecam result from its output directory

        :param directory: directory contains the possible facemap output files (`*.npy`, `*.pkl`, and `*.h5`)
        :param track_type: {'keypoints', 'pupil'}
        :param file_pattern: string prefix pattern to glob the facemap output file
        :return: :class:`FaceMapResult`
        """
        #
        try:
            svd_path = uglob(directory, f'{file_pattern}*.npy')
        except FileNotFoundError:
            svd = None
        else:
            svd = np.load(svd_path, allow_pickle=True).item()

        #
        try:
            meta_path = uglob(directory, f'{file_pattern}*.pkl')
        except FileNotFoundError:
            meta = None
            data = None
            with_keypoints = False
        else:
            with open(meta_path, 'rb') as f:
                meta = pickle.load(f)

            data_path = uglob(directory, f'{file_pattern}*.h5')
            data = h5py.File(data_path)['Facemap']

            with_keypoints = True

        cls._verify_data_content(directory, track_type, svd, data)

        return FaceMapResult(svd, meta, data, track_type, with_keypoints)

    @classmethod
    def _verify_data_content(cls, directory, track_type, svd, data):
        if track_type == 'keypoints' and data is None:
            raise RuntimeError(f'.h5 data file not found in the {directory}')
        elif track_type == 'pupil' and svd is None:
            raise RuntimeError(f'.npy svd file not found in the {directory}')

    @classmethod
    def launch_facemap_gui(cls, directory: PathLike,
                           with_keypoints: bool,
                           *,
                           file_pattern: str = '',
                           env_name: str = 'neuralib') -> None:
        """
        GUI view via cli.

        **Note that calling this method will overwrite `filenames` field in *proc.npy**

        :param directory: directory contains the possible facemap output files (`*.npy`, `*.pkl`, and `*.h5`),
            and also the raw video file
        :param with_keypoints: if has keypoint tracking result
        :param file_pattern: string prefix pattern to glob the facemap output file and raw avi file
        :param env_name: conda env name that installed the facemap package
        """
        import subprocess

        cls._modify_video_filenames_field(directory)

        svd_path = uglob(directory, f'{file_pattern}*.npy')

        cmds = ['conda', 'run', '-n', f'{env_name}', 'python', '-m', 'facemap']
        cmds.extend(CliArgs('--proc_npy', str(svd_path)).as_command())

        if with_keypoints:
            data_path = uglob(directory, '*.h5')
            cmds.extend(CliArgs('--keypoints', str(data_path)).as_command())

        fprint(f'{cmds=}')
        subprocess.check_call(cmds)

    @classmethod
    def _modify_video_filenames_field(cls,
                                      directory: PathLike, *,
                                      file_pattern: str = ''):
        """brute force rewrite ``filenames`` field in raw file"""
        svd_path = uglob(directory, f'{file_pattern}*.npy')
        video_path = uglob(directory, f'{file_pattern}*.avi')

        dat = np.load(svd_path, allow_pickle=True).item()
        dat['filenames'] = [[str(video_path)]]
        np.save(svd_path, dat, allow_pickle=True)
        fprint(f'overwrite filenames field to {str(video_path)}', vtype='warning')

    # ============== #
    # Pupil Tracking #
    # ============== #

    def get_pupil_tracking(self) -> PupilDict:
        """pupil tracking result"""
        ret = self.svd['pupil']
        if len(ret) == 1:
            return ret[0]
        raise NotImplementedError('')

    def get_pupil_area(self) -> np.ndarray:
        """pupil area (F, )"""
        return self.get_pupil_tracking()['area_smooth']

    def get_pupil_center_of_mass(self) -> np.ndarray:
        """center of mass of pupil tracking (F, 2)"""
        return self.get_pupil_tracking()['com']

    def get_pupil_location_movement(self) -> np.ndarray:
        """Calculate the Euclidean distance from the origin for each point in a 2D array"""
        com = self.get_pupil_center_of_mass()
        return np.sqrt(np.sum(com ** 2, axis=1))

    def get_eye_blink(self) -> np.ndarray:
        """eye blinking array (F, )"""
        ret = self.svd['blink']
        if len(ret) == 1:
            return ret[0]
        raise NotImplementedError('')

    # ========= #
    # Keypoints #
    # ========= #

    @property
    def keypoints(self) -> list[KeyPoint]:
        """list of keypoint name"""
        return list(self.data.keys())

    def __getitem__(self, keypoint: KeyPoint) -> KeyPointTrack:
        """get a specific keypoint tracking result"""
        if keypoint not in self.keypoints:
            raise KeyError(f'{keypoint} invalid, please select from {self.keypoints}')

        x = np.array(self.data[keypoint]['x'])
        y = np.array(self.data[keypoint]['y'])
        llh = np.array(self.data[keypoint]['likelihood'])

        return KeyPointTrack(keypoint, x, y, llh)

    def __iter__(self) -> Generator[KeyPointTrack, None, None]:
        """iterate all keypoints"""
        for kp in self.keypoints:
            yield self[kp]

    @overload
    def get(self, keypoint: KeyPoint) -> KeyPointTrack:
        """single keypoint"""
        pass

    @overload
    def get(self, keypoint: list[KeyPoint]) -> list[KeyPointTrack]:
        """multiple keypoints"""
        pass

    def get(self, keypoint):
        """get a single or multiple keypoint(s)"""
        if isinstance(keypoint, str):
            return self[keypoint]
        elif isinstance(keypoint, list):
            return [self[kp] for kp in keypoint]
        else:
            raise TypeError('')

    def as_array(self,
                 keypoint: list[KeyPoint] | KeyPoint | None = None,
                 with_outlier_filter: bool = True,
                 to_zscore: bool = True,
                 **kwargs) -> np.ndarray:
        """
        get keypoint(s) result as an 2D array with shape (K, F, 2). 3rd dim indicates the xy

        :param keypoint: keypoint
        :param with_outlier_filter:
        :param to_zscore:
        :param kwargs: pass through :meth:`~KeyPointTrack.with_outlier_filter()`
        :return: (K, F, 2)
        """
        if keypoint is not None:
            kps = self.get(keypoint)
        else:
            kps = [kp for kp in self]  # all

        #
        if not isinstance(kps, list):
            kps = [kps]

        ret = []
        for kp in kps:  # type: KeyPointTrack

            if with_outlier_filter:
                kp = kp.with_outlier_filter(**kwargs)

            if to_zscore:
                kp = kp.to_zscore()

            ret.append(np.vstack([kp.x, kp.y]).T)

        return np.array(ret)


# ================= #
# Individual Points #
# ================= #

@attrs.define
class KeyPointTrack:
    """single keypoint tracked result"""

    name: KeyPoint
    """name of keypoint"""
    x: np.ndarray
    """x loc (F,)"""
    y: np.ndarray
    """y loc (F,)"""
    likelihood: np.ndarray
    """tracking likelihood (F,)"""

    @property
    def mean_xy(self) -> np.ndarray:
        """mean x y loc"""
        return (self.x + self.y) / 2

    def with_outlier_filter(
            self,
            filter_window: int = 15,
            baseline_window: int = 50,
            max_spike: int = 25,
            max_diff: int = 25
    ) -> Self:
        """x,y with outlier filter (remove jump and do the interpolation)

        :param filter_window: window size for median filter
        :param baseline_window: window size for baseline estimation
        :param max_spike: maximum spike size
        :param max_diff: maximum difference between baseline and filtered signal
        :return: :class:`KeyPointTrack`

        """
        from facemap.utils import filter_outliers
        _x, _y = filter_outliers(self.x,
                                 self.y,
                                 filter_window,
                                 baseline_window,
                                 max_spike,
                                 max_diff)

        return attrs.evolve(self, x=_x, y=_y)

    def to_zscore(self) -> Self:
        """x, y z-scoring"""
        from scipy.stats import zscore
        _x = zscore(self.x)
        _y = zscore(self.y)
        return attrs.evolve(self, x=_x, y=_y)



from __future__ import annotations

import itertools

import numpy as np
from bokeh.palettes import Category20c
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colorbar import ColorbarBase
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

__all__ = [
    'ColorMapper',
    'get_customized_cmap',
    'ax_colorbar',
    'insert_colorbar'
]


class ColorMapper:  # TODO named as DiscreteColorMapper
    """map color to iterable object

    **Example of a `Dict` palette**

        >>> cmapper = ColorMapper(Category20c, 20)
        >>> x = ['1', '2', '3']
        >>> color_list = [cmapper[i] for i in x]

    **Example of a mpl `str` palette**

        >>> cmapper = ColorMapper('viridis', 20)
        >>> regions = ['rsc', 'vis', 'hpc']
        >>> color_list = [cmapper[r] for r in regions]
    """

    def __init__(self,
                 palette: dict[int, tuple] | str,
                 number: int):
        if isinstance(palette, dict):
            self._color_pool = iter(itertools.cycle(palette[number]))
        elif isinstance(palette, str):  # matplotlib
            import matplotlib as mpl
            mapper = mpl.colormaps[palette].resampled(number)
            self._color_pool = iter(itertools.cycle(mapper(range(number))))
        else:
            raise TypeError('')

        self._key_pool = {}

    def __getitem__(self, item: str):
        if item in self._key_pool:
            return self._key_pool[item]
        else:
            color = next(self._color_pool)
            self._key_pool[item] = color
            return color


def get_customized_cmap(name: str, value: tuple[float, float], numbers: int) -> np.ndarray:
    """
    Generate gradient color map array

    :param name: name of cmap
    :param value: value range, could be 0-1
    :param numbers: number of color (N)
    :return: (N, 4) RGBA
    """
    cmap = plt.get_cmap(name)
    return cmap(np.linspace(*value, numbers))


def ax_colorbar(ax: Axes, height="25%") -> Axes:
    return inset_axes(
        ax,
        width="5%",
        height=height,
        loc='upper left',
        bbox_to_anchor=(1.01, 0., 1, 1),
        bbox_transform=ax.transAxes,
        borderpad=0,
    )


def insert_colorbar(ax: Axes, im, **kwargs) -> ColorbarBase:
    cax = ax_colorbar(ax)

    return ax.figure.colorbar(im, cax=cax, **kwargs)



import dataclasses
from functools import cached_property
from pathlib import Path
from typing import Optional

import numpy as np
import polars as pl
from scipy.interpolate import interp1d
from typing_extensions import Self

from neuralib.argp import as_argument, argument
from neuralib.atlas.brainrender.core import BrainReconstructor
from neuralib.atlas.util import PLANE_TYPE
from neuralib.atlas.util import roi_points_converter
from neuralib.util.segement import grouped_iter

__all__ = ['ProbeReconstructor']


class ProbeReconstructor(BrainReconstructor):
    DESCRIPTION = 'For probe(s) track reconstruction'

    implant_depth: int = argument('-D', '--depth', required=True, help='implant depth in um')

    dye_label_only: bool = argument('--dye', help='only show the histology dye parts')

    csv_file: Path = as_argument(BrainReconstructor.csv_file).with_options(
        required=True,
        help='csv file after registration using 2dccf pipeline, point numbers equal to probe(s) * 2'
    )
    """
    Example::
    
        ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┐
        │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex │
        │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     │
        │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     │
        ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╡
        │ Primary visual area layer 6a      ┆ VISp6a  ┆ -3.81       ┆ 1.92        ┆ -3.12       ┆ 191     │
        │ optic radiation                   ┆ or      ┆ -4.08       ┆ 2.33        ┆ -3.12       ┆ 1217    │
        │ Posterolateral visual area layer… ┆ VISpl6a ┆ -4.28       ┆ 2.29        ┆ -3.12       ┆ 198     │
        │ Posterolateral visual area layer… ┆ VISpl5  ┆ -4.52       ┆ 2.17        ┆ -3.12       ┆ 197     │
        │ Subiculum                         ┆ SUB     ┆ -3.93       ┆ 4.36        ┆ -3.3        ┆ 536     │
        │ Entorhinal area medial part dors… ┆ ENTm5   ┆ -4.19       ┆ 4.39        ┆ -3.3        ┆ 515     │
        │ Entorhinal area medial part dors… ┆ ENTm2   ┆ -4.44       ┆ 4.39        ┆ -3.3        ┆ 510     │
        │ Entorhinal area medial part dors… ┆ ENTm1   ┆ -4.66       ┆ 4.29        ┆ -3.3        ┆ 509     │
        └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┘
    """

    plane_type: PLANE_TYPE = argument(
        '--plane-type', '-P',
        default='coronal',
        help='cutting orientation. Assume if multiple shanks were inserted along the AP axis, then do the sagittal'
             'slicing, if inserted along the ML axis, then do the coronal slicing '
    )

    raw: pl.DataFrame
    """raw csv file"""
    data: pl.DataFrame
    """sorted data based on plane_type"""

    def post_parsing(self):
        super().post_parsing()
        self.raw = pl.read_csv(self.csv_file)
        self.data = self.infer_probe_index(self.raw)

    def load(self):
        self.dye_label_only = True
        probe_dye = self.get_probe_object().shanks
        probe_dye = np.vstack(probe_dye)

        self.dye_label_only = False
        probe_theo = self.get_probe_object().with_theoretical().shanks
        probe_theo = np.vstack(probe_theo)

        self.add_points([probe_dye, probe_theo])

    @cached_property
    def number_shanks(self) -> int:
        return int(self.raw.shape[0] / 2)

    @dataclasses.dataclass
    class ShanksTrack:
        """shank object

        `Dimension parameters`:

            S = number of shanks

            P = number of sample points after interpolated reconstruction

        """

        rst: 'ProbeReconstructor'
        """`ProbeReconstructor`"""
        shanks: list[np.ndarray]
        """len: S, each shape: (P, 3) with ap, dv, ml"""

        def __post_init__(self):
            assert len(self.shanks) == self.rst.number_shanks

        def __len__(self) -> int:
            """S"""
            return len(self.shanks)

        def __getitem__(self, idx: int) -> np.ndarray:
            """get a shank array (P, 3)"""
            return self.shanks[idx]

        def with_theoretical(self, interval: int = 250) -> Self:
            """
            theoretical track based on implantation depth / angle

            :param interval: istance (um) relative to the specific shank. e.g., NeuroPixel 2.0 = 250 * x
            :return: :class:`ShanksTrack`
            """
            if not self.rst.dye_label_only:
                crop = self.rst.crop_outside_brain
                depth = self.rst.implant_depth
                ret = []
                s1 = self[0]

                for i in range(len(self)):
                    s_cur = self[i]
                    p = crop(s_cur, depth, dv_value=int(s1[0, 1]))

                    ret.append(p)
                    depth += _calc_shank_length_diff(s1, interval)
                    interval += interval

                return dataclasses.replace(self, shanks=ret)

            raise RuntimeError('')

    def get_probe_object(self) -> ShanksTrack:
        p = roi_points_converter(self.data)
        n_label_points = p.shape[0]

        ext_depth = None if self.dye_label_only else (0, 5000)

        fn = lambda p1, p2: self._shank_extend(np.array([p1, p2]), ext_depth=ext_depth)

        ret = []
        for (surface_idx, tip_idx) in grouped_iter(np.arange(n_label_points), 2):
            ret.append(fn(p[surface_idx], p[tip_idx]))

        return self.ShanksTrack(self, ret)

    def infer_probe_index(self, df: pl.DataFrame) -> pl.DataFrame:
        """
        probe in correct order index

        :param df: raw csv file
        :return: dataframe with `probe` and `probe_idx` fields.

        Example::

            ┌───────────────────────────────────┬─────────┬─────────────┬─────────────┬─────────────┬─────────┬───────────────┬───────────┐
            │ name                              ┆ acronym ┆ AP_location ┆ DV_location ┆ ML_location ┆ avIndex ┆ probe         ┆ probe_idx │
            │ ---                               ┆ ---     ┆ ---         ┆ ---         ┆ ---         ┆ ---     ┆ ---           ┆ ---       │
            │ str                               ┆ str     ┆ f64         ┆ f64         ┆ f64         ┆ i64     ┆ str           ┆ i64       │
            ╞═══════════════════════════════════╪═════════╪═════════════╪═════════════╪═════════════╪═════════╪═══════════════╪═══════════╡
            │ Primary visual area layer 6a      ┆ VISp6a  ┆ -3.81       ┆ 1.92        ┆ -3.12       ┆ 191     ┆ dorsal_label  ┆ 1         │
            │ Subiculum                         ┆ SUB     ┆ -3.93       ┆ 4.36        ┆ -3.3        ┆ 536     ┆ ventral_label ┆ 1         │
            │ optic radiation                   ┆ or      ┆ -4.08       ┆ 2.33        ┆ -3.12       ┆ 1217    ┆ dorsal_label  ┆ 2         │
            │ Entorhinal area medial part dors… ┆ ENTm5   ┆ -4.19       ┆ 4.39        ┆ -3.3        ┆ 515     ┆ ventral_label ┆ 2         │
            │ Posterolateral visual area layer… ┆ VISpl6a ┆ -4.28       ┆ 2.29        ┆ -3.12       ┆ 198     ┆ dorsal_label  ┆ 3         │
            │ Entorhinal area medial part dors… ┆ ENTm2   ┆ -4.44       ┆ 4.39        ┆ -3.3        ┆ 510     ┆ ventral_label ┆ 3         │
            │ Posterolateral visual area layer… ┆ VISpl5  ┆ -4.52       ┆ 2.17        ┆ -3.12       ┆ 197     ┆ dorsal_label  ┆ 4         │
            │ Entorhinal area medial part dors… ┆ ENTm1   ┆ -4.66       ┆ 4.29        ┆ -3.3        ┆ 509     ┆ ventral_label ┆ 4         │
            └───────────────────────────────────┴─────────┴─────────────┴─────────────┴─────────────┴─────────┴───────────────┴───────────┘

        """
        n_shanks = self.number_shanks

        df = (df.sort('DV_location')
              .with_columns(pl.Series(['dorsal_label'] * n_shanks + ['ventral_label'] * n_shanks).alias('probe')))

        if self.plane_type == 'sagittal':
            probe_order = 'AP_location'
        elif self.plane_type == 'coronal':
            probe_order = 'ML_location'
        else:
            raise RuntimeError('')

        df = (df.sort(by=['probe', probe_order], descending=[False, True])
              .with_columns(pl.Series(list(range(1, 1 + n_shanks)) * 2).alias('probe_idx'))
              .sort('probe_idx'))

        print(df)

        return df

    def isin_brain(self, shank: np.ndarray) -> np.ndarray:
        """
        determine if the probe points are in the brain

        :param shank: (N', 3)
        :return:
        """
        brain = self.get_atlas_brain_globe()

        ret = []
        for sh in shank:
            try:
                s = brain.structure_from_coords(sh, microns=True)
            except IndexError:
                s = 0
            ret.append(s != 0)

        return np.array(ret, dtype=bool)

    def crop_outside_brain(self, shank: np.ndarray,
                           distance: float,
                           dv_value: Optional[int] = None) -> np.ndarray:
        """
        crop the probe after doing the extension

        :param shank: (N', 3)
        :param distance: depth of insertion (might with an angle, in um). mostly records during the implantation
                    use the depth value that used while implantation to cutoff the bottom line.
        :param dv_value: if None, plot the probe if its in the brain
                if int type, plot the probe if dv larger than this value
        :return:
        """
        shank = shank[shank[:, 1] >= 0]

        if dv_value is None:
            m = self.isin_brain(shank)
        elif isinstance(dv_value, int):
            m = shank[:, 1] >= dv_value
        else:
            raise TypeError('')

        shank = shank[m]
        d = np.sqrt(np.sum((shank - shank[0]) ** 2, axis=1))

        return shank[d <= distance]

    @staticmethod
    def _shank_extend(shank: np.ndarray,
                      ext_depth: Optional[tuple[float, float]] = (0, 5000)) -> np.ndarray:
        """
        probe extension using extrapolation and interpolation

        :param shank: (2, 3), (points, (ap,dv,ml))
        :param ext_depth: depth in um, if None, only do the interpolation of the labelled points
        :return: extended shank
        """

        if ext_depth is not None:
            nn = np.arange(*ext_depth, 10)
        else:
            nn = np.arange(shank[0, 1], shank[-1, 1], 10)

        return interp1d(shank[:, 1], shank, axis=0, bounds_error=False, fill_value='extrapolate')(nn)


# ========= #


def _calc_shank_length_diff(shank: np.ndarray,
                            shank_interval: float) -> float:
    """
    use the vector of the probe, then calculate the unit vector

    :param shank: (N', 3)
    :param shank_interval: distance (um) relative to the specific shank. e.g., NeuroPixel 2.0 = 250 * x
    :return: unit vector
    """
    v = shank[-1] - shank[0]
    v = v / np.linalg.norm(v)  # unit vector

    # vector product: |n x v| = sin_theta |n| |v|
    # inline n value, got following formula
    sin_theta = np.linalg.norm(np.array([v[2], 0, -v[0]]))

    return shank_interval * sin_theta


if __name__ == '__main__':
    ProbeReconstructor().main()



from __future__ import annotations

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import Normalize

from neuralib.plot import plot_figure
from neuralib.plot.colormap import get_customized_cmap
from neuralib.util.util_type import ArrayLikeStr, PathLike
from neuralib.wrapper.facemap import FaceMapResult

__all__ = ['plot_facemap_keypoints',
           'plot_camp_time_series']


def plot_facemap_keypoints(fmap: FaceMapResult,
                           frame_interval: tuple[int, int],
                           keypoints: str | ArrayLikeStr | None = None,
                           outlier_filter: bool = True,
                           output: PathLike | None = None):
    """
    Plot all the keypoints

    :param fmap: :class:`~neuralib.wrapper.facemap.core.FaceMapResult`
    :param frame_interval: frame interval (start/end) for the plot
    :param keypoints: a keypoint name Or multiple keypoints. If None, then show all the keypoints
    :param outlier_filter: remove jump and do the interpolation
    :param output: output file. Show the fig if None
    :return:
    """
    if keypoints is None:
        kps = fmap.keypoints
    else:
        kps = list(keypoints)

    x = np.arange(*frame_interval)
    colors = get_customized_cmap('jet', (0, 1), len(kps))

    with plot_figure(output) as ax:
        for i, kp in enumerate(fmap.get(kps)):

            if outlier_filter:
                _kp = kp.with_outlier_filter()
            else:
                _kp = kp

            ax.plot(x, _kp.x[x], '-', color=colors[i], label=f'{_kp.name}_x')
            ax.plot(x, _kp.y[x], '--', color=colors[i], label=f'{_kp.name}_y')

        ax.set(xlabel='frame', ylabel='keypoint coordinates')
        ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)


def plot_camp_time_series(x: np.ndarray,
                          y: np.ndarray,
                          cmap: str = 'viridis',
                          output: PathLike | None = None,
                          **kwargs):
    """
    Plots a scatter plot with a colorbar

    :param x: X-axis values
    :param y: Y-axis values
    :param cmap: Colormap to use
    :param output: Output file path to save the plot
    :return:
    """
    if x.size != y.size:
        raise ValueError('size xy inconsistent')

    colors = get_customized_cmap(cmap, value=(0, 1), numbers=len(x))
    norm = Normalize(vmin=0, vmax=len(x))

    with plot_figure(output) as ax:
        ax.scatter(x, y, c=colors, alpha=0.5, **kwargs)
        cbar = plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax)
        cbar.set_label('frames')
        ax.set(xlabel='x', ylabel='y')



import unittest
from pathlib import Path
from typing import ClassVar

import polars as pl
import polars.testing

from neuralib.util.csv import csv_header


class TestCSVContextManager(unittest.TestCase):
    FILE: ClassVar[Path] = Path('.test.csv')

    def test_basic_usage(self):
        fields = ['neuron_id', 'variable_a', 'variable_b']
        total_neurons = 3
        val_a = [v * 0.5 for v in range(total_neurons)]
        val_b = [v * 2 for v in range(total_neurons)]

        with csv_header(self.FILE, fields) as csv:
            for neuron in range(total_neurons):
                csv(neuron, val_a[neuron], val_b[neuron])

        df = pl.read_csv(self.FILE)

        ans = pl.DataFrame(
            {
                f'{fields[0]}': pl.Series([i for i in range(total_neurons)]),
                f'{fields[1]}': pl.Series(val_a),
                f'{fields[2]}': pl.Series(val_b)
            }
        )

        pl.testing.assert_frame_equal(df, ans)

        self.FILE.unlink()

    def test_double_quotes_usage(self):
        fields = ['neuron_id', 'variable_a', 'variable_b']
        total_neurons = 3
        val_a = [v * 0.5 for v in range(total_neurons)]
        val_b = ['0.1,0.5,0.8', '0.7,0.7', '0.8,0.8']

        with csv_header(self.FILE, fields, quotes_header='variable_b') as csv:
            for neuron in range(total_neurons):
                csv(neuron, val_a[neuron], val_b[neuron])

        df = pl.read_csv(self.FILE)

        ans = pl.DataFrame(
            {
                f'{fields[0]}': pl.Series([i for i in range(total_neurons)]),
                f'{fields[1]}': pl.Series(val_a),
                f'{fields[2]}': pl.Series(val_b)
            }
        )

        pl.testing.assert_frame_equal(df, ans)

        self.FILE.unlink()


if __name__ == '__main__':
    unittest.main()



import sys

import numpy as np

__all__ = ['place_bayes']


def place_bayes(fr: np.ndarray,
                rate_map: np.ndarray,
                spatial_bin_size: float) -> np.ndarray:
    """
    Position decoding using population neuronal activity

    `Dimension parameters`:

        N = number of neurons

        T = number of temporal bins

        X = number of spatial bins


    .. seealso::

        https://github.com/buzsakilab/buzcode/blob/master/analysis/positionDecoding/placeBayes.m


    :param fr: firing rate 2D array. (T, N)
    :param rate_map: firing rate template. (X, N)
    :param spatial_bin_size: spatial bin size in cm

    :return: matrix of posterior probabilities. (T, X)
    """
    overflow = sys.float_info.min

    fr *= spatial_bin_size
    term2 = (-1) * spatial_bin_size * np.sum(rate_map, axis=1)
    term2 -= np.max(term2)  # normalize in log space
    term2 = np.exp(term2)  # (X,)

    if np.any(term2 < overflow):
        raise RuntimeError(f'OVERFLOW: {term2=}')

    outer_multiply = np.vectorize(np.multiply.outer, signature='(t),(s)->(t,s)')
    # outer_multiply((N, T), (N, X)) = (N, T, X)
    u = np.sum(outer_multiply(fr.T, np.log(rate_map.T)), axis=0)
    u -= np.max(u, axis=1, keepdims=True)  # normalize in log space
    u = np.exp(u)  # (T, X)

    if np.any(u < overflow):
        raise RuntimeError(f'OVERFLOW: {u=}')

    n_spatial_bins = rate_map.shape[0]
    pr = u * term2 / n_spatial_bins  # (T, X)

    pr /= np.sum(pr, axis=1, keepdims=True)

    if np.any(np.isnan(pr)):
        raise RuntimeError('pr approach the infinite')

    return pr



import numpy as np
from numba import jit

from neuralib.util.segement import Segment

__all__ = ['compute_velocity',
           'remove_jumps',
           'interpolate_gaps']


def compute_velocity(xy: np.ndarray, dx: float = 1.0) -> np.ndarray:
    """Compute velocity

    :param xy: (n,2) coordinates
    :param dx: time period between samples (for smoothing)
    :return:
        (n,) complex array
        velocity as a complex number. To compute speed, use np.abs(vel).
        To compute movement direction, use np.angle(vel).
    """
    return np.gradient(xy[:, 0], dx) + np.gradient(xy[:, 1], dx) * 1.0j


@jit(nopython=True)
def remove_jumps(t: np.ndarray,
                 xy: np.ndarray,
                 jump_size: float,
                 duration: float = 0.1) -> tuple[np.ndarray, list[list[int]]]:
    """

    :param t: (n, ) time vector
    :param xy: (n, 2) xy position coordinates
    :param jump_size: the maximum shift in position that is not yet considered a jump. Should be expressed in the same coordinates as xy
    :param duration: the maximum duration (in the same units as t) of a jump. If a jump lasts longer than maxduration, it will not be
        removed (as it may not actually be a jump)
    :return:
        xy : (n,2) array
            corrected x,y coordinated (positions during jumps are set to np.nan)
        jumps : sequence
            a list of [start,end] indices for each detected jump
    """
    xy = xy.copy()

    # only work with the valid coordinates
    valid = np.flatnonzero(np.logical_not(np.isnan(xy[:, 0])))

    if len(valid) < 2:
        return xy, [[np.int64(x) for x in range(0)]]

    last_valid = valid[0]
    found_jump = False
    jumps = [[np.int64(x) for x in range(0)]]

    # loop through all valid coordinates
    for idx in valid[1:]:

        # compute time since and distance to last valid coordinate
        dx = t[idx] - t[last_valid]
        dy = np.sqrt(np.sum((xy[idx] - xy[last_valid]) ** 2))

        if not found_jump and dy / (idx - last_valid) > jump_size:
            # we detected a jump larger than the threshold
            found_jump = True

        if found_jump and (dy <= jump_size or dx >= duration):
            # and now we detected a jump back within the maximum allowed duration
            # or this a long jump
            if dx < duration:
                xy[last_valid:idx, :] = np.NaN
                jumps.append([last_valid, idx])

            found_jump = False

        if not found_jump:
            last_valid = idx

    return xy, jumps


def interpolate_gaps(t: np.ndarray,
                     xy: np.ndarray,
                     interp_gap: float) -> np.ndarray:
    """Interpolate over gaps in position record.

    :param t: time vector
    :param xy: x,y coordinates
    :param interp_gap: maximum duration of a gap that should be interpolated over

    :return xy: (n,2) array
        corrected x,y coordinates
    """
    from scipy.interpolate import interp1d

    xy = xy.copy()

    # small gap interpolation
    invalid = np.isnan(xy[:, 0])
    valid = np.logical_not(invalid)
    # create index segments for all invalid gaps
    invalid_segments = Segment.from_logical(invalid)
    # select small gaps
    invalid_segments = invalid_segments[
        (t[invalid_segments.stop] - t[invalid_segments.start]) < interp_gap
        ]
    if len(invalid_segments) > 0:
        # construct index vector for all small gaps
        invalid_indices = np.concatenate(
            [np.arange(start, stop + 1) for start, stop in invalid_segments]
        )
        # interpolate small gaps
        fcn = interp1d(
            t[valid],
            xy[valid, :],
            kind="linear",
            bounds_error=False,
            axis=0,
            assume_sorted=True,
        )
        xy[invalid_indices, :] = fcn(t[invalid_indices])

    return xy



import contextlib
import functools
import textwrap
from typing import Union, ContextManager, Dict, Optional, List, Tuple

import pandas as pd
import polars as pl
import rich
from rich import box
from rich.console import Console, Capture
from rich.table import Table, Column

__all__ = [
    'Column',
    'Table',
    'rich_table',
    'rich_table_content',
    'rich_default_table',
    'rich_data_frame_table',
    'show_polars_header_types'
]


class TableLike:
    def __init__(self, table: Table, capture: Optional[Capture] = None):
        self.__table = table
        self.__capture = capture

    def add_column(self, header=None, footer=None, **kwargs):
        self.__table.add_column(header, footer, **kwargs)

    def __call__(self, *args):
        """

        :rtype: object
        """
        self.__table.add_row(*[str(it) if it is not None else '' for it in args])

    def get(self, prepend: Union[int, str] = None):
        ret = self.__capture.get()

        if prepend is not None:
            if isinstance(prepend, int):
                prepend = ' ' * prepend
            ret = textwrap.indent(ret, prepend)

        return ret


def _rich_table_header(*header: Union[int, str, Column]) -> List[Column]:
    def _header(ih: Tuple[int, Union[int, str, Column]]) -> Column:
        i, h = ih
        if isinstance(h, int):
            if i + 1 == len(header):  # last one
                return Column('', max_width=h)
            else:
                return Column('', min_width=h)
        elif isinstance(h, str):
            return Column(h)
        elif isinstance(h, Column):
            return h
        else:
            raise TypeError()

    return list(map(_header, enumerate(header)))


@functools.wraps(Table.__init__)
def rich_default_table(*header: Union[str, Column], **kwargs) -> Table:
    kwargs.setdefault('show_edge', False)
    kwargs.setdefault('box', box.SIMPLE)
    return Table(*_rich_table_header(*header), **kwargs)


@contextlib.contextmanager
@functools.wraps(Table.__init__)
def rich_table(*header: Union[int, str, Column],
               **kwargs) -> ContextManager[TableLike]:
    table = rich_default_table(*header, **kwargs)
    yield TableLike(table)
    rich.get_console().print(table)


@contextlib.contextmanager
@functools.wraps(Table.__init__)
def rich_table_content(*header: Union[int, str, Column],
                       **kwargs) -> ContextManager[TableLike]:
    console = Console()
    with console.capture() as capture:
        table = rich_default_table(*header, **kwargs)
        yield TableLike(table, capture)
        console.print(table)


def rich_data_frame_table(frame: Union[pd.DataFrame, pl.DataFrame, Dict], *,
                          show_dytpe=False) -> Table:
    if isinstance(frame, pd.DataFrame):
        return _rich_pandas_data_frame_table(frame, show_dytpe=show_dytpe)
    elif isinstance(frame, pl.DataFrame):
        return _rich_polars_data_frame_table(frame, show_dytpe=show_dytpe)
    else:
        return _rich_polars_data_frame_table(pl.DataFrame(frame), show_dytpe=show_dytpe)


def _rich_pandas_data_frame_table(frame: pd.DataFrame, *,
                                  show_dytpe=False) -> Table:
    table = rich_default_table()
    table.add_column(frame.index.name or '')
    for c in frame.columns:
        table.add_column(c)

    if show_dytpe:
        # TODO
        pass

    for c, d in frame.iterrows():
        table.add_row(str(c), *list(map(str, d)))

    rich.get_console().print(table)
    return table


def _rich_polars_data_frame_table(frame: pl.DataFrame, *,
                                  show_dytpe=False) -> Table:
    table = rich_default_table()
    for c in frame.columns:
        table.add_column(c)

    if show_dytpe:
        table.add_row(*list(map(str, frame.dtypes)))

    for row in frame.iter_rows():
        table.add_row(*list(map(str, row)))

    rich.get_console().print(table)
    return table


def show_polars_header_types(frame: pl.DataFrame):
    with pl.Config() as cfg:
        cfg.set_tbl_rows(0)
        cfg.set_tbl_hide_column_data_types(False)
        cfg.set_tbl_hide_column_names(False)
        cfg.set_tbl_hide_dtype_separator(True)
        cfg.set_tbl_hide_dataframe_shape(True)
        cfg.set_tbl_cols(-1)  # display all
        print(frame)



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import Any, final

import numpy as np
import tifffile

from neuralib.plot import plot_figure
from neuralib.plot.colormap import insert_colorbar
from neuralib.scanner import (
    AbstractConfocalScanner,
    parse_tif_meta,
    SceneIdx,
    DimCode,
    ZEISS_LSM_CHANNELS_ORDER,
    ZPROJ_TYPE
)
from neuralib.util.util_type import PathLike

__all__ = ['LSMConfocalScanner']


# TODO multiple scenes not impl.
# TODO sep plot func from the class
@final
@dataclasses.dataclass
class LSMConfocalScanner(AbstractConfocalScanner):
    """lsm confocal image data"""
    lsmfile: np.ndarray
    """lsm image array"""
    meta: dict[str, Any]

    @classmethod
    def load(cls, filepath: PathLike):
        lsm = tifffile.imread(filepath)
        return LSMConfocalScanner(lsm, cls.get_meta(filepath))

    @classmethod
    def get_meta(cls, filepath: PathLike) -> dict[str, Any]:
        return parse_tif_meta(Path(filepath), is_lsm=True)

    def get_dim_code(self) -> DimCode:
        return DimCode('ZCYX')

    @property
    def width(self) -> dict[SceneIdx, int]:
        return {0: self.lsmfile.shape[3]}

    @property
    def height(self) -> dict[SceneIdx, int]:
        return {0: self.lsmfile.shape[2]}

    @property
    def n_channels(self) -> dict[SceneIdx, int]:
        return {0: self.lsmfile.shape[1]}

    @property
    def n_zstacks(self) -> dict[SceneIdx, int]:
        return {0: self.lsmfile.shape[0]}

    def get_image(self, channel: int,
                  depth: int | slice | np.ndarray | None = None,
                  zproj_type: ZPROJ_TYPE = 'max',
                  norm: bool = True) -> np.ndarray:
        """

        :param channel:
        :param depth:
        :param zproj_type:
        :param norm:
        :return:
        """

        if depth is None:
            depth = np.arange(self.lsmfile.shape[0])

        img = self.lsmfile[depth, channel, :, :]

        if img.ndim == 3 and img.shape[0] != 1:
            img = self._zproj(img, zproj_type)

        if norm:
            img = img / np.max(img)

        return img

    def imshow(self, channel: int,
               depth: int | slice | np.ndarray | None = None,
               add_scale_bar: bool = True,
               zproj_type: ZPROJ_TYPE = 'max',
               norm: bool = True,
               output: PathLike | None = None):
        """

        :param channel:
        :param depth:
        :param add_scale_bar:
        :param zproj_type:
        :param norm:
        :param output:
        :return:
        """

        img = self.get_image(channel, depth, zproj_type, norm)

        with plot_figure(output) as ax:
            im = ax.imshow(img, cmap='Greys', vmax=0.5 if norm else None)
            insert_colorbar(ax, im)

            if add_scale_bar:
                # TODO
                pass

    def plot_merge_channel(self):
        with plot_figure(None) as ax:
            for c in range(self.lsmfile.shape[1]):
                img = self.get_image(c, depth=None)
                ax.imshow(img, cmap=ZEISS_LSM_CHANNELS_ORDER[c], alpha=0.8, vmax=0.3)



from typing import TypeVar

__all__ = ['check_attrs_in_clazz']

C = TypeVar('C')


def check_attrs_in_clazz(clazz: type[C], attr: str) -> bool:
    """
    Check if attr is in the class

    :param clazz: class name
    :param attr: attribute name
    :return:
    """
    if attr in getattr(clazz, '__annotations__', {}):
        return True

    for cls in clazz.mro()[1:]:  # Skip the first class as it's already checked
        if attr in getattr(cls, '__annotations__', {}):
            return True

    return False



from __future__ import annotations

from typing import Literal, NamedTuple

import numpy as np
from scipy.ndimage import gaussian_filter, filters

from neuralib.calimg.suite2p.core import Suite2PResult, SIGNAL_TYPE

__all__ = [
    'get_neuron_signal',
    'normalize_signal',
    'normalize_signal_factor',
    #
    'DFFSignal',
    'dff_signal',
    'calc_signal_baseline',
    #
    'sync_s2p_rigevent',
]

BASELINE_METHOD = Literal['maximin', 'constant', 'constant_prctile']


def get_neuron_signal(s2p: Suite2PResult,
                      n: int | np.ndarray | list[int],
                      signal_type: SIGNAL_TYPE = 'df_f',
                      normalize: bool = True,
                      dff: bool = True,
                      correct_neuropil: bool = True,
                      method: BASELINE_METHOD = 'maximin') -> tuple[np.ndarray, np.ndarray]:
    """
    Select neuronal signals for analysis. For single cell (F,) OR multiple cells (N, F)

    :param s2p: suite 2p result
    :param n: neuron index or index array
    :param signal_type: signal type. :data:`~neuralib.calimg.suite2p.core.SIGNAL_TYPE` {'df_f', 'spks'}
    :param normalize: 01 normalization for each neuron
    :param dff: normalize to the baseline fluorescence changed (dF/F)
    :param correct_neuropil: do the neuropil correction
    :param method: baseline calculation method {'maximin', 'constant', 'constant_prctile'}
    :return: tuple with (signal, baseline signal)
    """
    f = s2p.F[n]
    fneu = s2p.FNeu[n]

    if signal_type == 'df_f':

        if dff:
            s1 = dff_signal(f, fneu, s2p, correct_neuropil, method).dff
            s2 = np.full_like(s1, 0, dtype=int)
            if normalize:
                o, f = normalize_signal_factor(s1)
                s1 = (s1 - o) / f

        else:
            fcorr = f - 0.7 * fneu
            sig_baseline = s2p.signal_baseline
            window = int(s2p.window_baseline * s2p.fs)
            s1 = fcorr
            # baseline not a constant value, due to calcium decay or motion drift
            s2 = _maximin_filter(fcorr, sig_baseline, window)
            if normalize:
                o, f = normalize_signal_factor(s1)
                s1 = (s1 - o) / f
                ob, fb = normalize_signal_factor(s2)
                s2 = (s2 - ob) / fb

    elif signal_type == 'spks':
        s1 = s2p.spks[n]
        s2 = np.full_like(s1, 0, dtype=int)

        if normalize:
            s1 = normalize_signal(s1)

    else:
        raise ValueError('specify signal type either in df_f or spks')

    return s1, s2


def normalize_signal(s: np.ndarray, axis: int = -1) -> np.ndarray:
    """
    do 0 1 normalization

    :param s: signal
    :param axis: axis
    :return: normalized signal
    """
    o, f = normalize_signal_factor(s, axis)
    return (s - o) / f


def normalize_signal_factor(s: np.ndarray, axis: int = -1) -> tuple[np.ndarray, np.ndarray]:
    return np.min(s, axis=axis, keepdims=True), \
        np.max(s, axis=axis, keepdims=True) - np.min(s, axis=axis, keepdims=True)


# ============ #
# DF/F Compute #
# ============ #

class DFFSignal(NamedTuple):
    """Container for dF/F signal processing.

    `Dimension parameters`:

        N = number of neurons

        F = number of frame

    For single cell (F,) OR multiple cells (N, F)

    """

    s2p: Suite2PResult
    """Suite2PResult"""
    f: np.ndarray
    """fluorescence. (F,) | (N, F)"""
    fneu: np.ndarray
    """neuropil. (F,) | (N, F)"""
    fcorr: np.ndarray
    """f used for dff calculation, could be either corrected by fneu. (F,) | (N, F)"""
    f0: np.ndarray
    """f0. (F,) | (N, F)"""
    dff: np.ndarray
    """dff after f0 normalization. (F,) | (N, F)"""

    @property
    def dff_baseline(self) -> np.ndarray:
        """baseline of dff, supposed to be 0"""
        return np.full_like(self.dff, 0)

    @property
    def baseline_fluctuation(self) -> np.ndarray:
        """get the fluctuation of the fneu signal (F,) | (N, F).

        Perhaps not fully corrected with physiological reason.
        **used to get the baseline std (i.e., trial reliability metric)**
        """
        fneu_corr = 0.3 * self.fneu
        fneu_bas = calc_signal_baseline(fneu_corr, self.s2p, method='maximin')
        return 100 * ((fneu_corr - fneu_bas) / fneu_bas)

    def oasis_dcnv(self) -> np.ndarray:
        """spike deconvolution"""
        f = self.dff.astype(np.float32)  # (T,)
        v = w = l = s = np.zeros_like(f, dtype=np.float32)
        t = np.zeros_like(f, dtype=np.int64)
        _oasis_trace(f, v, w, t, l, self.s2p.indicator_tau, self.s2p.fs)
        return s


def dff_signal(f: np.ndarray,
               fneu: np.ndarray,
               s2p: Suite2PResult,
               correct_neuropil: bool = True,
               method: BASELINE_METHOD = 'maximin') -> DFFSignal:
    """
    df_f signal normalization container

    :param f: neuron signal (F)
    :param fneu: neuropil signal
    :param s2p: Suite2PResult
    :param correct_neuropil: whether do the subtraction using neuropil
    :param method:

    :return:
        DFFSignal
    """

    fcorr = f - 0.7 * fneu if correct_neuropil else f

    f0 = calc_signal_baseline(fcorr, s2p, method)
    dff = 100 * ((fcorr - f0) / f0)
    return DFFSignal(s2p, f, fneu, fcorr, f0, dff)


def calc_signal_baseline(signal: np.ndarray,
                         s2p: Suite2PResult,
                         method: BASELINE_METHOD = 'maximin') -> np.ndarray:
    """
    f0 calculation

    .. seealso:: source code from suite2p: ``suite2p.extraction.dcnv.preprocess``

    :param signal: signal activity. i.e., fcorr
    :param s2p: :class:`~neuralib.calimg.suite2p.core.Suite2PResult`
    :param method: BASELINE_METHOD
    :return:
        f0 baseline for the df/f calculation
    """

    if method == 'maximin':
        sig_baseline = s2p.signal_baseline
        window = int(s2p.window_baseline * s2p.fs)
        f0 = _maximin_filter(signal, sig_baseline, window)

    elif method == 'constant':
        signal = gaussian_filter(signal, [0., s2p.signal_baseline])
        f0 = np.min(signal, axis=-1)

    elif method == 'constant_prctile':
        sig_prc = np.percentile(signal, s2p.prctile_baseline, axis=-1)
        f0 = np.expand_dims(sig_prc, axis=-1)

    else:
        raise TypeError(f'unknown method{method}')

    return f0


def _maximin_filter(signal: np.ndarray, kernel: float, size: int) -> np.ndarray:
    """
    max/min method to calculate the signal baseline.
    **Note that this method is sensitive to the kernel size

    :param signal:
    :param kernel: standard deviation for Gaussian kernel
    :param size: length along which to calculate 1D minimum
    :return:
    """
    signal = filters.gaussian_filter1d(signal, kernel, axis=-1)
    signal = filters.minimum_filter1d(signal, size, axis=-1)
    baseline = filters.maximum_filter1d(signal, size, axis=-1)

    return baseline


def _oasis_trace(fcorr: np.ndarray,
                 v: np.ndarray,
                 w: np.ndarray,
                 t: np.ndarray,
                 l: np.ndarray,
                 tau: float,
                 fs: float) -> np.ndarray:
    """
    spike deconvolution on a single neuron

    :param fcorr: neuropil-subtracted, baseline-subtracted fluorescence (N,)
    :param v:
    :param w:
    :param t:
    :param l:
    :param tau: timescale of the sensor, used for the deconvolution kernel
    :param fs: sampling rate per plane
    :return:
        (F, ) spks
    """
    assert fcorr.ndim == 1
    total_frame = len(fcorr)
    g = -1. / (tau * fs)

    frame = 0
    ip = 0

    ret = np.zeros(total_frame, dtype=np.float32)

    while frame < total_frame:
        v[ip] = fcorr[frame]
        w[ip] = 1
        t[ip] = frame
        l[ip] = 1

        while ip > 0:
            if v[ip - 1] * np.exp(g * l[ip - 1]) > v[ip]:
                # violation of the constraint means merging pools
                f1 = np.exp(g * l[ip - 1])
                f2 = np.exp(2 * g * l[ip - 1])
                wnew = w[ip - 1] + w[ip] * f2
                v[ip - 1] = (v[ip - 1] * w[ip - 1] + v[ip] * w[ip] * f1) / wnew
                w[ip - 1] = wnew
                l[ip - 1] = l[ip - 1] + l[ip]
                ip -= 1
            else:
                break
        frame += 1
        ip += 1

    ret[t[1:ip]] = v[1:ip] - v[:ip - 1] * np.exp(g * l[:ip - 1])

    return ret


def _oasis(fcorr: np.ndarray,
           tau: float,
           fs: float) -> np.ndarray:
    """
    Computes non-negative deconvolution. no sparsity constraints

    .. seealso::

        suite2p.extraction.dcnv.preprocess

    :param fcorr: neuropil-subtracted, baseline-subtracted fluorescence, aka, fcorr - f0. (N, F)
    :param tau:  timescale of the sensor, used for the deconvolution kernel
    :param fs: sampling rate per plane
    :return:
        (N, T) spks
    """
    n_neurons, n_frames = fcorr.shape
    fcorr = fcorr.astype(np.float32)
    ret = np.zeros((n_neurons, n_frames), dtype=np.float32)

    for n in range(n_neurons):
        v = np.zeros((n_neurons, n_frames), dtype=np.float32)
        w = np.zeros((n_neurons, n_frames), dtype=np.float32)
        l = np.zeros((n_neurons, n_frames), dtype=np.float32)
        t = np.zeros((n_neurons, n_frames), dtype=np.int64)
        ret[n] = _oasis_trace(fcorr[n], v[n], w[n], t[n], l[n], tau, fs)

    return ret


# ================ #
# Sync Stimpy file #
# ================ #

def sync_s2p_rigevent(image_time: np.ndarray,
                      s2p: Suite2PResult,
                      plane: int | None = 0) -> np.ndarray:
    """
    Check if the registered frame number is consistent in .riglog,
    then do the alignment to make the shape the same

    * ASSUME both recordings are sync

    * ASSUME sequential scanning pattern. i.e., 0, 1, 2, 3 ETL order

    :param image_time: riglog image event
    :param s2p: :class:`~neuralib.calimg.suite2p.core.Suite2PResult`
    :param plane: number of optical plane
    :return: image_time: with same len as s2p results
    """
    n_plane = s2p.n_plane
    image_fr = s2p.fs

    if plane is not None:
        image_time = image_time[plane::n_plane]
    else:
        # TODO used npy in combined file
        raise NotImplementedError('')

    n_image = len(image_time)
    n_frame = s2p.n_frame
    n_diff = abs(n_image - n_frame)

    if n_diff >= n_frame / 100:
        raise RuntimeError('too large difference between riglog imaging header and calcium data')

    if n_image > n_frame:
        # drop latest, mostly this case
        image_time = image_time[:n_frame]
    elif n_image < n_frame:
        # padding at the tail
        image_time = np.hstack((image_time, [image_time[-1] + (i + 1) / image_fr for i in range(n_diff)]))

    return image_time



from __future__ import annotations

import abc
from pathlib import Path
from typing import NamedTuple, Final, Any, Literal, get_args, Iterable

import attrs
import numpy as np
import polars as pl
from scipy.io import loadmat
from scipy.io.matlab import MatlabOpaque

from neuralib.atlas.type import HEMISPHERE_TYPE
from neuralib.atlas.util import PLANE_TYPE
from neuralib.atlas.view import SlicePlane, load_slice_view
from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint
from neuralib.util.utils import uglob, joinn

__all__ = [
    'AbstractCCFDir',
    'CoronalCCFDir',
    'CoronalCCFOverlapDir',
    'SagittalCCFDir',
    'SagittalCCFOverlapDir',
    #
    'CCFTransMatrix',
    'load_transform_matrix'
]

# ======================== #
# CCF DataFolder Structure #
# ======================== #

CCF_GLOB_TYPE = Literal[
    'zproj',
    'roi',
    'roi_cpose',
    'resize',
    'resize_overlap',
    'processed',
    'transformation_matrix',
    'transformation_img',
    'transformation_img_overlap'
]

CHANNEL_SUFFIX = Literal['r', 'g', 'b', 'merge', 'overlap']


class AbstractCCFDir(metaclass=abc.ABCMeta):
    """
    ::

        ANIMAL_001/ (root)
            ├── raw/ (optional)
            ├── zproj/
            │    └── ANIMAL_001_g*_s*_{channel}.tif
            ├── roi/
            │    └── ANIMAL_001_g*_s*_{channel}.roi
            ├── roi_cpose/
            │    └── ANIMAL_001_g*_s*_{channel}.roi
            ├── resize/ (src for the allenccf, if sagittal slice, could be resize_contra and resize_ipsi)
            │    ├── ANIMAL_001_g*_s*_resize.tif
            │    └── processed/
            │           ├── ANIMAL_001_g*_s*_resize_processed.tif
            │           └── transformations/
            │                 ├── ANIMAL_001_g*_s*_resize_processed_transformed.tif
            │                 ├── ANIMAL_001_g*_s*_resize_processed_transform_data.mat
            │                 └── labelled_regions/
            │                       ├── {*channel}_roitable.csv
            │                       └── parsed_data /
            │                             └── parsed_csv_merge.csv
            ├── resize_*_overlap/ (optional, same structure as **resize**, for dual channels labeling)
            │
            └── output_files/ (for generate output fig)

    """

    def __new__(
            cls,
            root: PathLike,
            with_overlap_sources: bool = True,
            plane_type: PLANE_TYPE = 'coronal',
            hemisphere_type: HEMISPHERE_TYPE | None = None,
            auto_mkdir: bool = True,
    ):

        if plane_type == 'coronal':
            if with_overlap_sources:
                return object.__new__(CoronalCCFOverlapDir)
            else:
                return object.__new__(CoronalCCFDir)

        elif plane_type == 'sagittal':

            if hemisphere_type is None or hemisphere_type not in get_args(HEMISPHERE_TYPE):
                raise ValueError(f'invalid hemisphere_type for sagittal dir: {hemisphere_type}')

            #
            if with_overlap_sources:
                return object.__new__(SagittalCCFOverlapDir)
            else:
                return object.__new__(SagittalCCFDir)

        else:
            raise ValueError(f'invalid plane type: {plane_type}')

    def __init__(
            self,
            root: PathLike,
            with_overlap_sources: bool = True,
            plane_type: PLANE_TYPE = 'coronal',
            hemisphere_type: HEMISPHERE_TYPE | None = None,
            auto_mkdir: bool = True,
    ):
        r"""

        :param root: root path (i.e., \*/ANIMAL_001)
        :param with_overlap_sources:
        :param plane_type:
        :param hemisphere_type:
        :param auto_mkdir:
        """
        self.root: Final[Path] = root
        self.with_overlap_sources = with_overlap_sources
        self.plane_type: Final[PLANE_TYPE] = plane_type
        self.hemisphere: HEMISPHERE_TYPE | None = hemisphere_type if plane_type == 'sagittal' else None

        if auto_mkdir:
            self._init_folder_structure()

    def __len__(self):
        """number of slices"""
        return len(list(self.resize_folder.glob('*.tif')))

    def __iter__(self):
        return self.resize_folder.glob('*.tif')

    @abc.abstractmethod
    def _init_folder_structure(self) -> None:
        pass

    slice_name: str = None  # assign when glob

    @abc.abstractmethod
    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path:
        """

        :param glass_id: Glass slide number
        :param slice_id: Slice sequencing under a glass slide
            (i.e., zigzag from upper left -> bottom left ->-> bottom right)
        :param glob_type:
        :param hemisphere:
        :param channel:
        :return:
        """
        pass

    @property
    @abc.abstractmethod
    def resize_folder(self) -> Path:
        pass

    # ============================ #
    # Default Dir for the pipeline #
    # ============================ #

    @property
    def default_iter_dir(self) -> Iterable[Path]:
        return [
            self.raw_folder,
            self.resize_folder,
            self.roi_folder,
            self.cpose_roi_folder,
            self.zproj_folder,
            self.processed_folder,
            self.transformed_folder,
            self.labelled_roi_folder,
            self.parsed_data_folder,
            self.output_folder
        ]

    @property
    def animal(self) -> str:
        return self.root.name

    @property
    def raw_folder(self) -> Path:
        return self.root / 'raw'

    @property
    def roi_folder(self) -> Path:
        return self.root / 'roi'

    @property
    def cpose_roi_folder(self) -> Path:
        return self.root / 'roi_cpose'

    @property
    def zproj_folder(self) -> Path:
        return self.root / 'zproj'

    # =============================== #
    # Dual Channel Overlap (Optional) #
    # =============================== #

    @property
    def resize_overlap_folder(self) -> Path | None:
        if not self.with_overlap_sources:
            raise ValueError('')
        return

    @property
    def processed_folder_overlap(self) -> Path | None:
        if not self.with_overlap_sources:
            raise ValueError('')
        return

    @property
    def transformed_folder_overlap(self) -> Path | None:
        if not self.with_overlap_sources:
            raise ValueError('')
        return

    @property
    def labelled_roi_folder_overlap(self) -> Path | None:
        if not self.with_overlap_sources:
            raise ValueError('')
        return

    # ========================================== #
    # CCF folder (MATLAB pipeline auto-generate) #
    # ========================================== #

    @property
    def processed_folder(self) -> Path:
        return self.resize_folder / 'processed'

    @property
    def transformed_folder(self) -> Path:
        return self.processed_folder / 'transformations'

    @property
    def labelled_roi_folder(self) -> Path:
        return self.transformed_folder / 'labelled_regions'

    @property
    def parsed_data_folder(self) -> Path:
        return self.labelled_roi_folder / 'parsed_data'

    @property
    def parse_csv(self) -> Path:
        return self.parsed_data_folder / 'parsed_csv_merge.csv'

    # ======= #
    # Outputs #
    # ======= #

    @property
    def output_folder(self) -> Path:
        return self.root / 'output_files'

    def figure_output(self, *suffix, sep='_') -> Path:
        ret = self.output_folder / joinn(sep, *suffix)
        return ret.with_suffix('.pdf')

    def csv_output(self, *suffix, sep='_') -> Path:
        ret = self.output_folder / joinn(sep, *suffix)
        return ret.with_suffix('.csv')

    @property
    def roi_atlas_output(self) -> Path:
        return self.output_folder / 'roiatlas'

    @property
    def roi_atlas_ibl_output(self) -> Path:
        return self.output_folder / 'roiatlas_ibl'

    # ========= #
    # File Glob #
    # ========= #

    def get_slice_id(self, glass_id: int,
                     slice_id: int,
                     hemisphere: Literal['i', 'c'] | None = None,
                     channel: CHANNEL_SUFFIX | None = None) -> str:

        ret = f'{self.animal}_{glass_id}_{slice_id}'
        if hemisphere is not None:
            ret += f'_{hemisphere}'
        if channel is not None:
            ret += f'_{channel}'
        return ret

    def get_transformation_matrix(self,
                                  glass_id: int,
                                  slice_id: int,
                                  plane_type: PLANE_TYPE) -> CCFTransMatrix:
        return load_transform_matrix(
            self.glob(glass_id, slice_id, 'transformation_matrix'),
            plane_type
        )


# ============= #
# Coronal Slice #
# ============= #

class CoronalCCFDir(AbstractCCFDir):
    """Base folder structure for 2dccf pipeline"""

    def _init_folder_structure(self):
        for d in self.default_iter_dir:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    @property
    def resize_folder(self) -> Path:
        return self.root / 'resize'

    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE, *,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        if glob_type in ('roi', 'roi_cpose', 'zproj'):
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere, channel=channel)
        else:
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere)

        #
        if glob_type == 'roi':
            return uglob(self.roi_folder, f'{name}*.roi')
        elif glob_type == 'roi_cpose':
            return uglob(self.cpose_roi_folder, f'{name}*cpose.roi')
        elif glob_type == 'zproj':
            return uglob(self.zproj_folder, f'{name}.*')
        elif glob_type == 'resize':
            return uglob(self.zproj_folder, f'{name}_resize.*')
        elif glob_type == 'transformation_matrix':
            return uglob(self.transformed_folder, f'{name}*.mat')
        elif glob_type == 'transformation_img':
            return uglob(self.transformed_folder, f'{name}*.tif')
        else:
            return


class CoronalCCFOverlapDir(CoronalCCFDir):
    """ 2dccf Folder structure for multiple sources overlap labeling
    For example: Dual tracing with different fluorescence protein, and tend to see the overlap channel counts"""

    def _init_folder_structure(self):
        super()._init_folder_structure()

        iter_overlap = (
            self.resize_overlap_folder,
            self.processed_folder,
            self.transformed_folder_overlap
        )
        for d in iter_overlap:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    @property
    def resize_overlap_folder(self) -> Path:
        """for double labeling channel
        since maximal 2 channels for roi detection,
        then need extra folder use pseudo-color
        """
        return self.root / 'resize_overlap'

    @property
    def processed_folder_overlap(self) -> Path:
        return self.resize_overlap_folder / 'processed'

    @property
    def transformed_folder_overlap(self) -> Path:
        return self.processed_folder_overlap / 'transformations'

    @property
    def labelled_roi_folder_overlap(self) -> Path:
        return self.transformed_folder_overlap / 'labelled_regions'

    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE, *,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        ret = super().glob(glass_id, slice_id, glob_type, hemisphere=hemisphere, channel=channel)

        if ret is None:
            if glob_type == 'resize_overlap':
                return uglob(self.resize_overlap_folder, f'{self.slice_name}_resize_overlap.*')
            elif glob_type == 'transformation_img_overlap':
                return uglob(self.transformed_folder_overlap, f'{self.slice_name}*.tif')

        return


# ============== #
# Sagittal Slice #
# ============== #

class SagittalCCFDir(AbstractCCFDir):

    def _init_folder_structure(self) -> None:
        for d in self.default_iter_dir:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    def glob(self, glass_id: int, slice_id: int, glob_type: CCF_GLOB_TYPE, hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        if glob_type in ('roi', 'roi_cpose', 'zproj'):
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere, channel=channel)
        else:
            self.slice_name = name = self.get_slice_id(glass_id, slice_id, hemisphere=hemisphere)

            #
        if glob_type == 'roi':
            return uglob(self.roi_folder, f'{name}*.roi')
        elif glob_type == 'roi_cpose':
            return uglob(self.cpose_roi_folder, f'{name}*cpose.roi')
        elif glob_type == 'zproj':
            return uglob(self.zproj_folder, f'{name}.*')
        elif glob_type == 'resize':
            return uglob(self.zproj_folder, f'{name}_resize.*')
        elif glob_type == 'transformation_matrix':
            return uglob(self.transformed_folder, f'{name}*.mat')
        elif glob_type == 'transformation_img':
            return uglob(self.transformed_folder, f'{name}*.tif')
        else:
            return

    @property
    def resize_folder(self) -> Path:
        if self.hemisphere == 'ipsi':
            return self.root / 'resize_ipsi'
        elif self.hemisphere == 'contra':
            return self.root / 'resize_contra'
        elif self.hemisphere == 'both':
            return self.root / 'resize'
        else:
            raise ValueError('')


class SagittalCCFOverlapDir(SagittalCCFDir):
    def _init_folder_structure(self):
        super()._init_folder_structure()

        iter_overlap = (
            self.resize_overlap_folder,
            self.processed_folder,
            self.transformed_folder_overlap
        )
        for d in iter_overlap:
            if not d.exists():
                d.mkdir(parents=True, exist_ok=True)
                fprint(f'auto make folder <{d.name}> for {self.animal}', vtype='io')

    @property
    def resize_overlap_folder(self) -> Path:
        """for double labeling channel
        since maximal 2 channels for roi detection,
        then need extra folder use pseudo-color
        """
        if self.hemisphere == 'ipsi':
            return self.root / 'resize_ipsi_overlap'
        elif self.hemisphere == 'contra':
            return self.root / 'resize_contra_overlap'
        elif self.hemisphere == 'both':
            return self.root / 'resize_overlap'
        else:
            raise ValueError('')

    @property
    def processed_folder_overlap(self) -> Path:
        return self.resize_overlap_folder / 'processed'

    @property
    def transformed_folder_overlap(self) -> Path:
        return self.processed_folder_overlap / 'transformations'

    @property
    def labelled_roi_folder_overlap(self) -> Path:
        return self.transformed_folder_overlap / 'labelled_regions'

    def glob(self,
             glass_id: int,
             slice_id: int,
             glob_type: CCF_GLOB_TYPE, *,
             hemisphere: Literal['i', 'c'] | None = None,
             channel: CHANNEL_SUFFIX | None = None) -> Path | None:

        ret = super().glob(glass_id, slice_id, glob_type, hemisphere=hemisphere, channel=channel)

        if ret is None:
            if glob_type == 'resize_overlap':
                return uglob(self.resize_overlap_folder, f'{self.slice_name}_resize_overlap.*')
            elif glob_type == 'transformation_img_overlap':
                return uglob(self.transformed_folder_overlap, f'{self.slice_name}*.tif')

        return


# ===================== #
# Transformation matrix #
# ===================== #

class MatMatrix(NamedTuple):
    allen_location: np.ndarray
    """abnormal parsing in loadmat. [996 array([  0, -12], dtype=int16)]"""
    transform: MatlabOpaque
    """mat meta info. currently not used"""
    transform_points: np.ndarray
    """[(P, 2), (P, 2)] transformation point for 2dccf GUI site view"""

    @property
    def slice_index(self) -> int:
        return int(self.allen_location[0])

    # noinspection PyTypeChecker
    @property
    def delta_values(self) -> tuple[int, int]:
        """shifting values for yx axis (height, width) for a specific plane type.

        Coronal slice: [0]:DV, [1]:ML

        Sagittal Slice: [0]:DV, [1]:AP

        """
        delta = self.allen_location[1]
        return delta[0], delta[1]


def load_transform_matrix(filepath: PathLike,
                          plane_type: PLANE_TYPE,
                          resolution: int = 10,
                          default_name: str = 'test') -> CCFTransMatrix:
    """
    matrix for image transformation

    :param filepath: transformation matrix .mat file
    :param plane_type: ``PLANE_TYPE`` {'coronal', 'sagittal', 'transverse'}
    :param resolution: atlas resolution
    :param default_name: name of the ``CCFTransMatrix``
    :return: ``CCFTransMatrix``
    """
    mat = loadmat(filepath, squeeze_me=True, struct_as_record=False)['save_transform']
    f = Path(filepath).name

    try:
        name = f[:f.index('_resize')]
    except ValueError:
        name = default_name

    return CCFTransMatrix(name,
                          MatMatrix(
                              mat.allen_location,
                              mat.transform,
                              mat.transform_points
                          ),
                          plane_type,
                          resolution=resolution)


@attrs.frozen
class CCFTransMatrix:
    slice_id: str
    matrix: MatMatrix
    plane_type: PLANE_TYPE
    resolution: int = attrs.field(kw_only=True, default=10)

    @property
    def slice_index(self) -> int:
        return self.matrix.slice_index

    @property
    def delta_xy(self) -> tuple[int, int]:
        """map delta value to xy slice view"""
        return self.matrix.delta_values[1], self.matrix.delta_values[0]

    def get_slice_plane(self) -> SlicePlane:
        ret = (
            load_slice_view('ccf_template', self.plane_type, allen_annotation_res=self.resolution)
            .plane_at(self.slice_index)
        )

        dw = self.delta_xy[0]
        dh = self.delta_xy[1]

        ret = ret.with_offset(
            dw=dw + 1 if dw != 0 else 0,  # avoid index err
            dh=dh + 1 if dh != 0 else 0,  # avoid index err
        )

        return ret

    def matrix_info(self, to_polars: bool = True) -> dict[str, Any] | pl.DataFrame:
        ret = {
            'slice_id': self.slice_id,
            'plane_type': self.plane_type,
            'atlas_resolution': self.resolution,
            'slice_index': self.slice_index,
            'reference_value': self.get_slice_plane().reference_value,
            'dw': self.delta_xy[0],
            'dh': self.delta_xy[1]
        }

        if to_polars:
            ret = pl.DataFrame(ret)

        return ret



from __future__ import annotations

import datetime
import operator
import re
from pathlib import Path
from typing import TypeVar, Iterable, overload, TYPE_CHECKING, Any

from neuralib.util.table import rich_table
from .expr import SqlExpr, SqlField

if TYPE_CHECKING:
    from .stat import Cursor

__all__ = [
    'str_to_datetime',
    'datetime_to_str',
    'take',
    'infer_eq',
    'infer_cmp',
    'infer_in',
    'resolve_field_type',
    'cast_to_sql',
    'cast_from_sql',
    'get_fields_from_schema',
    'map_foreign',
    'pull_foreign',
    'rich_sql_table'
]

T = TypeVar('T')
V = TypeVar('V')


def str_to_datetime(t: str) -> datetime.datetime:
    return datetime.datetime.strptime(t, '%Y-%m-%d %H:%M:%S')


def datetime_to_str(t: datetime.datetime) -> str:
    return t.strftime('%Y-%m-%d %H:%M:%S')


@overload
def take(index: int, coll: Cursor | Iterable[tuple[T, ...]]) -> list[T]:
    pass


@overload
def take(index: type[V], coll: Cursor | Iterable[tuple[T, ...]]) -> list[V]:
    pass


@overload
def take(index: tuple[int, ...], coll: Cursor | Iterable[tuple[T, ...]]) -> list[tuple[T, ...]]:
    pass


@overload
def take(index: tuple[V], coll: Cursor | Iterable[T]) -> list[tuple[V]]:
    pass


@overload
def take(index: V, coll: Cursor | Iterable[T]) -> list[V]:
    pass


def take(index, coll: Cursor | Iterable):
    """
    A help function that compose itemgetter and mapping functions.

    >>> @named_tuple_table_class
    ... class A:
    ...     a: int
    ...     b: str
    >>> take(0, [(0, 'a'), (1, 'b')])
    [0, 1]
    >>> take(A.a, [A(0, 'a'), A(1, 'b')])
    [0, 1]


    :param index:
    :param coll:
    :return:
    """
    if isinstance(index, type):
        return list(map(lambda it: index(*it), coll))

    if isinstance(index, int):
        return list(map(operator.itemgetter(index), coll))

    if isinstance(index, tuple) and all([isinstance(it, int) for it in index]):
        def _index(item):
            return tuple([item[it] for it in index])

        return list(map(_index, coll))

    from .table import table_field_names
    from .stat import Cursor

    if isinstance(index, SqlField):
        if isinstance(coll, Cursor):
            index = coll.headers.index(index.field.name)
        else:
            fields = table_field_names(index.field.table)
            index = fields.index(index.field.name)

        return list(map(operator.itemgetter(index), coll))

    if isinstance(index, tuple):
        index = cast(tuple[SqlField], index)
        if isinstance(coll, Cursor):
            index = tuple([coll.headers.index(it.field.name) for it in index])
        else:
            fields = table_field_names(index[0].field.table)
            index = tuple([fields.index(it.field.name) for it in index])

        return take(index, coll)

    raise TypeError()


def infer_eq(x: T, v: T | str, *, prepend: str = '', append: str = '') -> SqlExpr | None:
    """
    A help function to make a SQL ``=`` expression.

    >>> infer_eq(A.a, 1) # doctest: SKIP
    A.a = 1
    >>> infer_eq(A.a, '!1') # doctest: SKIP
    A.a != 1
    >>> infer_eq(A.a, '1%') # doctest: SKIP
    A.a LIKE '1%'

    :param x:
    :param v:
    :param prepend:
    :param append:
    :return:
    """
    if not isinstance(x, SqlField):
        raise TypeError()

    if v is None:
        return None
    if isinstance(v, SqlExpr):
        return v

    invert = False
    if isinstance(v, str) and v.startswith('!'):
        invert = True
        v = v[1:]

    if isinstance(v, str) and '%' in v:
        from .func_stat import like, not_like
        return not_like(x, v) if invert else like(x, v)

    if isinstance(v, str) and (prepend == '%' or append == '%'):
        from .func_stat import like, not_like
        v = prepend + v + append
        return not_like(x, v) if invert else like(x, v)

    return x != v if invert else x == v


def infer_cmp(x: T, v: T | str | range | slice) -> SqlExpr | None:
    """
    A help function to make a SQL comparison expression.

    >>> infer_cmp(A.a, range(0, 10))  # doctest: SKIP
    A.a BETWEEN 0 AND 9
    >>> infer_cmp(A.a, slice(0, 10))  # doctest: SKIP
    A.a BETWEEN 0 AND 10
    >>> infer_cmp(A.a, '<10')  # doctest: SKIP
    A.a < 10
    >>> infer_cmp(A.a, 10)  # doctest: SKIP
    A.a = 10

    :param x:
    :param v:
    :return:
    """
    if not isinstance(x, SqlField):
        raise TypeError()

    if v is None:
        return None
    if isinstance(v, SqlExpr):
        return v
    if isinstance(v, (int, float)):
        return x == v
    if isinstance(v, (range, slice)):
        return infer_in(x, v)

    if '%' in v:
        from .func_stat import like
        return like(x, v)

    if v.startswith('<='):
        return x <= float(v[2:])
    elif v.startswith('>='):
        return x >= float(v[2:])
    elif v.startswith('<'):
        return x < float(v[1:])
    elif v.startswith('>'):
        return x > float(v[1:])

    invert = False
    if v.startswith('!'):
        invert = True
        v = v[1:]

    return x != v if invert else x == v


def infer_in(x: T, v: T | str | list[str] | slice | range) -> SqlExpr | None:
    """
    A help function to make a SQL containing expression.

    >>> infer_in(A.a, '1')  # doctest: SKIP
    A.a == '1'
    >>> infer_in(A.a, range(0, 10))  # doctest: SKIP
    A.a BETWEEN 0 AND 9
    >>> infer_in(A.a, ['a', 'b'])  # doctest: SKIP
    A.a IN ('a', 'b')

    :param x:
    :param v:
    :return:
    """
    if not isinstance(x, SqlField):
        raise TypeError()

    if v is None:
        return None
    if isinstance(v, SqlExpr):
        return v

    if isinstance(v, str):
        return infer_eq(x, v)
    if isinstance(v, (range, slice)):
        return x.between(v)
    return x.contains(v)


def resolve_field_type(f_type: type) -> tuple[type, type, bool]:
    """

    SQL primary types:

    * bool: BOOLEAN
    * int: INT
    * float: FLOAT
    * str: TEXT
    * bytes: BLOB
    * datetime.date: DATETIME
    * datetime.datetime: DATETIME

    Python type mapping

    * `T|None`: `resolve_field_type(T)` null-able
    * `T|V` : supported not
    * `Literal`: `str`
    * `Path`: `str`

    :param f_type:
    :return: (raw_type, sql_type, not_null)
    """
    import typing

    sql_type = f_type
    o = typing.get_origin(f_type)

    if o == typing.Annotated:
        return resolve_field_type(typing.get_args(f_type)[0])

    elif o == typing.Union:
        a = typing.get_args(f_type)
        if len(a) == 2:
            try:
                i = a.index(type(None))
            except ValueError as e:
                raise RuntimeError('Union type is not supported now') from e
            else:
                f_type = a[1 - i]
                _, sql_type, _ = resolve_field_type(f_type)
                return f_type, sql_type, False

    elif o == typing.Literal:
        return str, str, True

    elif f_type == Path:
        return f_type, str, True

    return f_type, sql_type, True


def cast_to_sql(raw_type: type[T], sql_type: type[V], value: T) -> V:
    if value is None:
        return None
    if sql_type == str:
        return str(value)
    if raw_type == Any:
        return value
    return value


def cast_from_sql(raw_type: type[T], sql_type: type[V], value: V) -> T:
    if value is None:
        return None
    if raw_type == Any:
        return value
    if sql_type in (int, float, bool):
        return value
    if raw_type == datetime.datetime and isinstance(value, str):
        return datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
    if raw_type == datetime.date and isinstance(value, str):
        return datetime.datetime.strptime(value, '%Y-%m-%d').date()
    if raw_type == sql_type:
        return value
    if raw_type == Path:
        return Path(value)
    if callable(raw_type):
        return raw_type(value)
    return value


def get_fields_from_schema(schema: str) -> list[str]:
    schema = schema[schema.index('(') + 1:]
    try:
        schema = schema[:schema.rfind(')')]
    except ValueError:
        pass

    schema = re.sub(r'\(.+?\)', '', schema.strip())

    found = []
    for field in schema.split(','):
        field = field.strip().split(' ')
        if field[0] in ('FOREIGN', 'UNIQUE', 'PRIMARY', 'CHECK'):
            break

        field = field[0]
        if len(field):
            if field.startswith('[') and field.endswith(']'):
                field = field[1:-1]
            found.append(field)

    return found


def map_foreign(value: T, foreign: type[V]) -> Cursor[V]:
    """
    Let a table ``T`` with a foreign constraint refer to table ``V``,
    map a ``T`` data to the ``V`` data.

    :param value:
    :param foreign: a foreign constraint
    :return:
    """
    from .table import table_foreign_field
    from .stat_start import select_from

    table = type(value)
    if (constraint := table_foreign_field(table, foreign)) is None:
        raise RuntimeError(f'not a foreign constraint : {foreign}')

    # SELECT * FROM V
    # WHERE AND*([V.field == t.field for field in constraint])
    return select_from(constraint.foreign_table).where(*[
        getattr(constraint.foreign_table, f) == getattr(value, t)
        for (t, f) in zip(constraint.fields, constraint.foreign_fields)
    ]).submit()


def pull_foreign(target: type[T], foreign: V) -> Cursor[T]:
    """
    Let a table ``T`` with a foreign constraint refer to table ``V``,
    pull ``T`` data from a ``V`` data.

    :param target: target table ``T``
    :param foreign: a foreign data ``V`` referred to.
    :return:
    """
    from .table import table_foreign_field
    from .stat_start import select_from

    if (constraint := table_foreign_field(target, type(foreign))) is None:
        raise RuntimeError(f'not a foreign constraint')

    # SELECT * FROM T
    # WHERE AND*([T.field == v.field for field in constraint])
    return select_from(constraint.table).where(*[
        getattr(constraint.table, t) == getattr(foreign, f)
        for (t, f) in zip(constraint.fields, constraint.foreign_fields)
    ]).submit()


def rich_sql_table(table: type[T], value: list[T]):
    from .table import table_field_names
    with rich_table(*table_field_names(table)) as _table:
        for _value in value:
            _table(*_value)
